# LIBRARIES
from django.db import models, connections, connection as default_connection
from django.db.models.sql.datastructures import EmptyResultSet
from django.db.models.query import Q
from google.appengine.api import datastore

# DJANGAE
from djangae.db.backends.appengine.query import transform_query, Query, WhereNode
from djangae.test import TestCase


DEFAULT_NAMESPACE = default_connection.ops.connection.settings_dict.get("NAMESPACE")


class TransformTestModel(models.Model):
    field1 = models.CharField(max_length=255)
    field2 = models.CharField(max_length=255, unique=True)
    field3 = models.CharField(null=True, max_length=255)
    field4 = models.TextField()

    class Meta:
        app_label = "djangae"

class InheritedModel(TransformTestModel):
    class Meta:
        app_label = "djangae"

class TransformQueryTest(TestCase):

    def test_polymodel_filter_applied(self):
        query = transform_query(
            connections['default'],
            InheritedModel.objects.filter(field1="One").all().query
        )
        query.prepare()

        self.assertEqual(2, len(query.where.children))
        self.assertTrue(query.where.children[0].children[0].is_leaf)
        self.assertTrue(query.where.children[1].children[0].is_leaf)
        self.assertEqual("class", query.where.children[0].children[0].column)
        self.assertEqual("field1", query.where.children[1].children[0].column)

    def test_basic_query(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.all().query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [ TransformTestModel._meta.db_table ])
        self.assertIsNone(query.where)

    def test_and_filter(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.filter(field1="One", field2="Two").all().query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [ TransformTestModel._meta.db_table ])
        self.assertTrue(query.where)
        self.assertEqual(2, len(query.where.children)) # Two child nodes

    def test_exclude_filter(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.exclude(field1="One").all().query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [ TransformTestModel._meta.db_table ])
        self.assertTrue(query.where)
        self.assertEqual(1, len(query.where.children)) # One child node
        self.assertTrue(query.where.children[0].negated)
        self.assertEqual(1, len(query.where.children[0].children))

    def test_ordering(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.filter(field1="One", field2="Two").order_by("field1", "-field2").query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [ TransformTestModel._meta.db_table ])
        self.assertTrue(query.where)
        self.assertEqual(2, len(query.where.children)) # Two child nodes
        self.assertEqual(["field1", "-field2"], query.order_by)

    def test_projection(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.only("field1").query
        )

        self.assertItemsEqual(["id", "field1"], query.columns)

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.values_list("field1").query
        )

        self.assertEqual(set(["field1"]), query.columns)

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.defer("field1", "field4").query
        )

        self.assertItemsEqual(set(["id", "field2", "field3"]), query.columns)

    def test_no_results_returns_emptyresultset(self):
        self.assertRaises(
            EmptyResultSet,
            transform_query,
            connections['default'],
            TransformTestModel.objects.none().query
        )

    def test_offset_and_limit(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.all()[5:10].query
        )

        self.assertEqual(5, query.low_mark)
        self.assertEqual(10, query.high_mark)

    def test_isnull(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.filter(field3__isnull=True).all()[5:10].query
        )

        self.assertTrue(query.where.children[0].value)
        self.assertEqual("ISNULL", query.where.children[0].operator)

    def test_distinct(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.distinct("field2", "field3").query
        )

        self.assertTrue(query.distinct)
        self.assertEqual(query.columns, set(["field2", "field3"]))

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.distinct().values("field2", "field3").query
        )

        self.assertTrue(query.distinct)
        self.assertEqual(query.columns, set(["field2", "field3"]))

    def test_order_by_pk(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("pk").query
        )

        self.assertEqual("__key__", query.order_by[0])

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("-pk").query
        )

        self.assertEqual("-__key__", query.order_by[0])

    def test_reversed_ordering(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("pk").reverse().query
        )

        self.assertEqual("-__key__", query.order_by[0])

    def test_clear_ordering(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("pk").order_by().query
        )

        self.assertFalse(query.order_by)

    def test_projection_on_textfield_disabled(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.values_list("field4").query
        )

        self.assertFalse(query.columns)
        self.assertFalse(query.projection_possible)


from djangae.tests.test_connector import Relation
from djangae.db.backends.appengine.dnf import normalize_query


class QueryNormalizationTests(TestCase):
    """
        The parse_dnf function takes a Django where tree, and converts it
        into a tree of one of the following forms:

        [ (column, operator, value), (column, operator, value) ] <- AND only query
        [ [(column, operator, value)], [(column, operator, value) ]] <- OR query, of multiple ANDs
    """

    def test_and_with_child_or_promoted(self):
        from .test_connector import TestUser
        """
            Given the following tree:

                   AND
                  / | \
                 A  B OR
                      / \
                     C   D

             The OR should be promoted, so the resulting tree is

                      OR
                     /   \
                   AND   AND
                  / | \ / | \
                 A  B C A B D
        """


        query = Query(TestUser, "SELECT")
        query.where = WhereNode()
        query.where.children.append(WhereNode())
        query.where.children[-1].column = "A"
        query.where.children[-1].operator = "="
        query.where.children.append(WhereNode())
        query.where.children[-1].column = "B"
        query.where.children[-1].operator = "="
        query.where.children.append(WhereNode())
        query.where.children[-1].connector = "OR"
        query.where.children[-1].children.append(WhereNode())
        query.where.children[-1].children[-1].column = "C"
        query.where.children[-1].children[-1].operator = "="
        query.where.children[-1].children.append(WhereNode())
        query.where.children[-1].children[-1].column = "D"
        query.where.children[-1].children[-1].operator = "="

        query = normalize_query(query)

        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(2, len(query.where.children))
        self.assertFalse(query.where.children[0].is_leaf)
        self.assertFalse(query.where.children[1].is_leaf)
        self.assertEqual(query.where.children[0].connector, "AND")
        self.assertEqual(query.where.children[1].connector, "AND")
        self.assertEqual(3, len(query.where.children[0].children))
        self.assertEqual(3, len(query.where.children[1].children))

    def test_and_queries(self):
        from .test_connector import TestUser
        qs = TestUser.objects.filter(username="test").all()

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertTrue(1, len(query.where.children))
        self.assertEqual(query.where.children[0].children[0].column, "username")
        self.assertEqual(query.where.children[0].children[0].operator, "=")
        self.assertEqual(query.where.children[0].children[0].value, "test")

        qs = TestUser.objects.filter(username="test", email="test@example.com")

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertTrue(2, len(query.where.children[0].children))
        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(query.where.children[0].connector, "AND")
        self.assertEqual(query.where.children[0].children[0].column, "username")
        self.assertEqual(query.where.children[0].children[0].operator, "=")
        self.assertEqual(query.where.children[0].children[0].value, "test")
        self.assertEqual(query.where.children[0].children[1].column, "email")
        self.assertEqual(query.where.children[0].children[1].operator, "=")
        self.assertEqual(query.where.children[0].children[1].value, "test@example.com")

        qs = TestUser.objects.filter(username="test").exclude(email="test@example.com")
        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))


        self.assertTrue(2, len(query.where.children[0].children))
        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(query.where.children[0].connector, "AND")
        self.assertEqual(query.where.children[0].children[0].column, "username")
        self.assertEqual(query.where.children[0].children[0].operator, "=")
        self.assertEqual(query.where.children[0].children[0].value, "test")
        self.assertEqual(query.where.children[0].children[1].column, "email")
        self.assertEqual(query.where.children[0].children[1].operator, "<")
        self.assertEqual(query.where.children[0].children[1].value, "test@example.com")
        self.assertEqual(query.where.children[1].children[0].column, "username")
        self.assertEqual(query.where.children[1].children[0].operator, "=")
        self.assertEqual(query.where.children[1].children[0].value, "test")
        self.assertEqual(query.where.children[1].children[1].column, "email")
        self.assertEqual(query.where.children[1].children[1].operator, ">")
        self.assertEqual(query.where.children[1].children[1].value, "test@example.com")


        instance = Relation(pk=1)
        qs = instance.related_set.filter(headline__startswith='Fir')

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertTrue(2, len(query.where.children[0].children))
        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(query.where.children[0].connector, "AND")
        self.assertEqual(query.where.children[0].children[0].column, "relation_id")
        self.assertEqual(query.where.children[0].children[0].operator, "=")
        self.assertEqual(query.where.children[0].children[0].value, 1)
        self.assertEqual(query.where.children[0].children[1].column, "_idx_startswith_headline")
        self.assertEqual(query.where.children[0].children[1].operator, "=")
        self.assertEqual(query.where.children[0].children[1].value, u"Fir")


    def test_or_queries(self):
        from .test_connector import TestUser
        qs = TestUser.objects.filter(
            username="python").filter(
            Q(username__in=["ruby", "jruby"]) | (Q(username="php") & ~Q(username="perl"))
        )

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        # After IN and != explosion, we have...
        # (AND: (username='python', OR: (username='ruby', username='jruby', AND: (username='php', AND: (username < 'perl', username > 'perl')))))

        # Working backwards,
        # AND: (username < 'perl', username > 'perl') can't be simplified
        # AND: (username='php', AND: (username < 'perl', username > 'perl')) can become (OR: (AND: username = 'php', username < 'perl'), (AND: username='php', username > 'perl'))
        # OR: (username='ruby', username='jruby', (OR: (AND: username = 'php', username < 'perl'), (AND: username='php', username > 'perl')) can't be simplified
        # (AND: (username='python', OR: (username='ruby', username='jruby', (OR: (AND: username = 'php', username < 'perl'), (AND: username='php', username > 'perl'))
        # becomes...
        # (OR: (AND: username='python', username = 'ruby'), (AND: username='python', username='jruby'), (AND: username='python', username='php', username < 'perl') \
        #      (AND: username='python', username='php', username > 'perl')

        self.assertTrue(4, len(query.where.children[0].children))

        self.assertEqual(query.where.children[0].connector, "AND")
        self.assertEqual(query.where.children[0].children[0].column, "username")
        self.assertEqual(query.where.children[0].children[0].operator, "=")
        self.assertEqual(query.where.children[0].children[0].value, "python")
        self.assertEqual(query.where.children[0].children[1].column, "username")
        self.assertEqual(query.where.children[0].children[1].operator, "=")
        self.assertEqual(query.where.children[0].children[1].value, "php")
        self.assertEqual(query.where.children[0].children[2].column, "username")
        self.assertEqual(query.where.children[0].children[2].operator, "<")
        self.assertEqual(query.where.children[0].children[2].value, "perl")

        self.assertEqual(query.where.children[1].connector, "AND")
        self.assertEqual(query.where.children[1].children[0].column, "username")
        self.assertEqual(query.where.children[1].children[0].operator, "=")
        self.assertEqual(query.where.children[1].children[0].value, "python")
        self.assertEqual(query.where.children[1].children[1].column, "username")
        self.assertEqual(query.where.children[1].children[1].operator, "=")
        self.assertEqual(query.where.children[1].children[1].value, "jruby")

        self.assertEqual(query.where.children[2].connector, "AND")
        self.assertEqual(query.where.children[2].children[0].column, "username")
        self.assertEqual(query.where.children[2].children[0].operator, "=")
        self.assertEqual(query.where.children[2].children[0].value, "python")
        self.assertEqual(query.where.children[2].children[1].column, "username")
        self.assertEqual(query.where.children[2].children[1].operator, "=")
        self.assertEqual(query.where.children[2].children[1].value, "php")
        self.assertEqual(query.where.children[2].children[2].column, "username")
        self.assertEqual(query.where.children[2].children[2].operator, ">")
        self.assertEqual(query.where.children[2].children[2].value, "perl")

        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(query.where.children[3].connector, "AND")
        self.assertEqual(query.where.children[3].children[0].column, "username")
        self.assertEqual(query.where.children[3].children[0].operator, "=")
        self.assertEqual(query.where.children[3].children[0].value, "python")
        self.assertEqual(query.where.children[3].children[1].column, "username")
        self.assertEqual(query.where.children[3].children[1].operator, "=")
        self.assertEqual(query.where.children[3].children[1].value, "ruby")

        qs = TestUser.objects.filter(username="test") | TestUser.objects.filter(username="cheese")

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(2, len(query.where.children))
        self.assertTrue(query.where.children[0].is_leaf)
        self.assertEqual("cheese", query.where.children[0].value)
        self.assertTrue(query.where.children[1].is_leaf)
        self.assertEqual("test", query.where.children[1].value)

        qs = TestUser.objects.using("default").filter(username__in=set()).values_list('email')

        with self.assertRaises(EmptyResultSet):
            query = normalize_query(transform_query(
                connections['default'],
                qs.query
            ))

        qs = TestUser.objects.filter(username__startswith='Hello') |  TestUser.objects.filter(username__startswith='Goodbye')

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertEqual(2, len(query.where.children))
        self.assertEqual("_idx_startswith_username", query.where.children[0].column)
        self.assertEqual(u"Goodbye", query.where.children[0].value)
        self.assertEqual("_idx_startswith_username", query.where.children[1].column)
        self.assertEqual(u"Hello", query.where.children[1].value)


        qs = TestUser.objects.filter(pk__in=[1, 2, 3])
        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertEqual(3, len(query.where.children))
        self.assertEqual("__key__", query.where.children[0].column)
        self.assertEqual("__key__", query.where.children[1].column)
        self.assertEqual("__key__", query.where.children[2].column)
        self.assertEqual({
                datastore.Key.from_path(TestUser._meta.db_table, 1, namespace=DEFAULT_NAMESPACE),
                datastore.Key.from_path(TestUser._meta.db_table, 2, namespace=DEFAULT_NAMESPACE),
                datastore.Key.from_path(TestUser._meta.db_table, 3, namespace=DEFAULT_NAMESPACE),
            }, {
                query.where.children[0].value,
                query.where.children[1].value,
                query.where.children[2].value,
            }
        )

        qs = TestUser.objects.filter(pk__in=[1, 2, 3]).filter(username="test")
        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertEqual(3, len(query.where.children))
        self.assertEqual("__key__", query.where.children[0].children[0].column)
        self.assertEqual("test", query.where.children[0].children[1].value)

        self.assertEqual("__key__", query.where.children[1].children[0].column)
        self.assertEqual("test", query.where.children[0].children[1].value)

        self.assertEqual("__key__", query.where.children[2].children[0].column)
        self.assertEqual("test", query.where.children[0].children[1].value)

        self.assertEqual({
                datastore.Key.from_path(TestUser._meta.db_table, 1, namespace=DEFAULT_NAMESPACE),
                datastore.Key.from_path(TestUser._meta.db_table, 2, namespace=DEFAULT_NAMESPACE),
                datastore.Key.from_path(TestUser._meta.db_table, 3, namespace=DEFAULT_NAMESPACE),
            }, {
                query.where.children[0].children[0].value,
                query.where.children[1].children[0].value,
                query.where.children[2].children[0].value,
            }
        )



from mapreduce import input_readers
import itertools
import logging
from django.apps import apps
from djangae import utils
from djangae.storage import UniversalNewLineBlobReader


logger = logging.getLogger(__name__)


class DjangoInputReader(input_readers.InputReader):

    REQUIRED_PARAMS = ('model',)

    def __init__(self, start_id, end_id, model, *args, **kwargs):
        self.shard_id = 1
        self.start_id = start_id
        self.end_id = end_id
        self.raw_model = model
        self.db = kwargs.pop('db', 'default')
        app, model = self.raw_model.split('.')
        self.model = apps.get_model(app, model)
        super(DjangoInputReader, self).__init__(*args, **kwargs)

    def __iter__(self):
        if self.start_id > self.end_id:
            # This can happen if we are the last shard and
            # the shard size caused each previous shard to process an additional model
            return

        query = self.model.objects.using(self.db)

        if self.start_id:
            query = query.filter(pk__gt=self.start_id).filter(pk__lte=self.end_id)
        query = query.order_by('pk')

        for model in utils.get_in_batches(query, batch_size=500):
            # From the mapreduce docs (AbstractDatastoreInputReader):
            #     The caller must consume yielded values so advancing the KeyRange
            #     before yielding is safe.
            self.start_id = model.id
            yield model

    @classmethod
    def validate(cls, mapper_spec):
        if mapper_spec.input_reader_class() != cls:
            raise input_readers.BadReaderParamsError("Input reader class mismatch")
        params = input_readers._get_params(mapper_spec)
        for param in cls.REQUIRED_PARAMS:
            if not param in params:
                raise input_readers.BadReaderParamsError("Parameter missing: %s" % param)

    @classmethod
    def split_input(cls, mapper_spec):
        shard_count = mapper_spec.shard_count

        # Grab the input parameters for the split
        params = input_readers._get_params(mapper_spec)
        logger.info("Params: %r", params)

        db = params['db']
        # Unpickle the query
        app, model = params['model'].split('.')
        model = apps.get_model(app, model)

        # Grab the lowest pk
        query = model.objects.using(db).all()
        query = query.order_by('pk')

        try:
            first_id = query.values_list('pk', flat=True)[:1][0]

            query = query.order_by('-pk')
            last_id = query.values_list('pk', flat=True)[:1][0]
        except IndexError:
            return [DjangoInputReader(0, 0, params['model'], db=db)]

        pk_range = last_id - first_id

        logger.info("Query range: %s - %s = %s", first_id, last_id, pk_range)

        if pk_range < shard_count or shard_count == 1:
            return [DjangoInputReader(first_id-1, last_id, params['model'], db=db)]

        readers = []
        max_shard_size = int(float(pk_range) / float(shard_count))
        if pk_range % shard_count:
            max_shard_size += 1

        shard_id = 1
        # Splitting could be much smarter by taking a __scatter__ sample and
        # clustering, which is how the DatastoreInputWriter from the mapreduce
        # splits on pks
        for i in itertools.count(first_id-1, max_shard_size):
            if i >= last_id:
                break

            shard_start_id = i
            shard_end_id = i + max_shard_size
            if shard_end_id > last_id:
                shard_end_id = last_id

            logger.info("Creating shard: %s - %s", shard_start_id, shard_end_id)
            reader = DjangoInputReader(shard_start_id, shard_end_id, params['model'], db=db)
            reader.shard_id = shard_id
            readers.append(reader)
            shard_id += 1
        return readers

    @classmethod
    def from_json(cls, input_shard_state):
        start_id = input_shard_state['start']
        end_id = input_shard_state['end']
        shard_id = input_shard_state['shard_id']
        model = input_shard_state['model']
        db = input_shard_state['db']

        reader = DjangoInputReader(start_id, end_id, model, db=db)
        reader.shard_id = shard_id
        return reader

    def to_json(self):
        return {
            'start': self.start_id,
            'end': self.end_id,
            'shard_id': self.shard_id,
            'model': self.raw_model,
            'db': self.db,
        }


class DjangoQuerySpec(object):
  """Encapsulates everything about a query needed by DatastoreInputReader."""

  DEFAULT_BATCH_SIZE = 50

  def __init__(self,
               entity_kind,
               keys_only=None,
               filters=None,
               batch_size=None,
               model_class_path=None,
               app=None,
               ns=None):
    self.entity_kind = entity_kind
    self.keys_only = keys_only or False
    self.filters = filters or None
    self.batch_size = batch_size or self.DEFAULT_BATCH_SIZE
    self.model_class_path = model_class_path
    self.app = app
    self.ns = ns

  def to_json(self):
    return {"entity_kind": self.entity_kind,
            "keys_only": self.keys_only,
            "filters": self.filters,
            "batch_size": self.batch_size,
            "model_class_path": self.model_class_path,
            "app": self.app,
            "ns": self.ns}

  @classmethod
  def from_json(cls, json):
    return cls(json["entity_kind"],
               json["keys_only"],
               json["filters"],
               json["batch_size"],
               json["model_class_path"],
               json["app"],
               json["ns"])


class BlobstoreUniversalLineInputReader(input_readers.BlobstoreLineInputReader):
    """A version of the BlobstoreLineInputReader that works with Mac, Windows and Linux line endings"""

    def __init__(self, blob_key, start_position, end_position):
        """Initializes this instance with the given blob key and character range.

        This BlobstoreInputReader will read from the first record starting after
        strictly after start_position until the first record ending at or after
        end_position (exclusive). As an exception, if start_position is 0, then
        this InputReader starts reading at the first record.

        Args:
          blob_key: the BlobKey that this input reader is processing.
          start_position: the position to start reading at.
          end_position: a position in the last record to read.
        """
        self._blob_key = blob_key
        self._blob_reader = UniversalNewLineBlobReader(blob_key,
            self._BLOB_BUFFER_SIZE,
            start_position)
        self._end_position = end_position
        self._has_iterated = False
        self._read_before_start = bool(start_position)



from google.appengine.ext import deferred

from djangae.test import TestCase, _get_queued_tasks


def my_task():
    """
    Basic task for testing task queues.
    """
    pass


class TaskQueueTests(TestCase):

    def test_get_queued_tasks_flush(self):
        deferred.defer(my_task)
        deferred.defer(my_task, _queue='another')

        # We don't use self.assertNumTasksEquals here because we want to flush.
        tasks = _get_queued_tasks(self.taskqueue_stub, queue_name='default')
        self.assertEqual(1, len(tasks))

        tasks = _get_queued_tasks(self.taskqueue_stub, queue_name='another')
        self.assertEqual(1, len(tasks))

        tasks = _get_queued_tasks(self.taskqueue_stub)
        self.assertEqual(0, len(tasks))



from google.appengine.api.memcache import *


def set_wrapper(func):
    """
        Apparently App Engine's memcache.set can't deal with a negative time, but some Django
        session tests do this intentionally, so we just wrap set to ignore calls with a negative
        time
    """
    def _wrapped(self, key, value, time=0, min_compress_len=0, namespace=None):
        if time < 0:
            return False

        return func(self, key, value, time, min_compress_len, namespace)

    return _wrapped

Client.set = set_wrapper(Client.set)



import threading
import new
import logging

import django
from django.db import router, connections, models
from django.apps import apps
from django.utils.encoding import smart_text
from djangae.crc64 import CRC64


logger = logging.getLogger(__name__)

class ListQS(list):
    """
        Hacky wrapper around a list to satisfy related_descriptors.py when accessing contenttypes through
        foreign keys
    """
    def get(self):
        return self[0]


class SimulatedContentTypeManager(models.Manager):
    """
        Simulates content types without actually hitting the datastore.

        The methods in this class are literally stubs which allow the normal Django
        code (Django admin) to believe it's talking to a database.

        ONLY A SMALL SUBSET OF CONTENTTYPE QUERIES ARE SUPPORTED AND THOSE UNSUPPORTED
        MAY JUST RETURN THE WRONG RESULTSET RATHER THAN THROWING AN ERROR. YOU HAVE BEEN WARNED! :)
    """
    use_for_related_fields = True

    _store = threading.local()

    def __init__(self, model=None, *args, **kwargs):
        super(SimulatedContentTypeManager, self).__init__(*args, **kwargs)
        self.model = model

    def _get_model(self):
        """ If we're in a migration, then the 'fake' model class will be passed
            into __init__   and we'll use that.  Otherwise we'll use the 'real'
            ContentType class.
        """
        from django.contrib.contenttypes.models import ContentType
        return self.model or ContentType

    def _get_id(self, app_label, model):
        """
            Generate a unique (ish) ID given the app label and model name
        """
        crc = CRC64()
        crc.append(app_label)
        crc.append(model)
        return crc.fini() - (2 ** 63)  # GAE integers are signed so we shift down

    def _get_opts(self, model, for_concrete_model):
        if for_concrete_model:
            model = model._meta.concrete_model
        elif getattr(model, '_deferred', False) or model is getattr(models, 'DEFERRED', None):
            model = model._meta.proxy_for_model
        return model._meta

    def _update_queries(self, models):
        """
            This is just to satisfy the contenttypes tests which check that queries are executed at certain
            times. It's a bit hacky but it works for that purpose.
        """
        ContentType = self._get_model()
        conn = connections[router.db_for_write(ContentType)]

        if getattr(conn, "use_debug_cursor", getattr(conn, "force_debug_cursor", False)):
            for model in models or []:
                if model not in self._store.queried_models:
                    conn.queries.append("select * from {}".format(ContentType._meta.db_table))
                    break

        if not hasattr(self._store, "queried_models"):
            self._store.queried_models = set()

        self._store.queried_models |= set(models or [])

    def _repopulate_if_necessary(self, models=None):
        if not hasattr(self._store, "queried_models"):
            self._store.queried_models = set()

        if not hasattr(self._store, "constructed_instances"):
            self._store.constructed_instances = {}

        self._update_queries(models)

        if not hasattr(self._store, "content_types"):
            all_models = [(x._meta.app_label, x._meta.model_name, x) for x in apps.get_models()]

            self._update_queries([(x[0], x[1]) for x in all_models])

            content_types = {}
            for app_label, model_name, model in all_models:
                content_type_id = self._get_id(app_label, model_name)

                content_types[content_type_id] = {
                    "id": content_type_id,
                    "app_label": app_label,
                    "model": model_name,
                }
                if django.VERSION < (1, 9):
                    content_types[content_type_id]["name"] = smart_text(model._meta.verbose_name_raw)

            self._store.content_types = content_types

    def get_by_natural_key(self, app_label, model):
        self._repopulate_if_necessary(models=[(app_label, model)])
        return self.get(id=self._get_id(app_label, model))

    def get_for_model(self, model, for_concrete_model=True):
        opts = self._get_opts(model, for_concrete_model)
        self._repopulate_if_necessary(models=[(opts.app_label, opts.model_name)])
        return self.get_by_natural_key(opts.app_label, opts.model_name)

    def get_for_models(self, *models, **kwargs):
        for_concrete_model = kwargs.get("for_concrete_models", True)

        self._update_queries(
            [(self._get_opts(x, for_concrete_model).app_label, self._get_opts(x, for_concrete_model).model_name) for x in models]
        )
        ret = {}
        for model in models:
            ret[model] = self.get_for_model(model, for_concrete_model)

        return ret

    def get_for_id(self, id):
        return self.get(pk=id)

    def clear_cache(self):
        self._store.queried_models = set()

    def _get_from_store(self, id):
        try:
            return self._store.content_types[id]
        except KeyError:
            ContentType = self._get_model()
            raise ContentType.DoesNotExist()

    def get(self, *args, **kwargs):
        # Special case for Django >= 1.11 which sometimes generates Q() objects
        # with a single filter when querying content types. This handles that one
        # case.. if we find there are others we'll have to do something less hacky!
        if args and not kwargs:
            if len(args) == 1 and len(args[0].children) == 1 and not args[0].negated:
                kwargs = dict(args[0].children)
            else:
                raise ValueError("Unsupported Q operation")

        ContentType = self._get_model()
        self._repopulate_if_necessary()

        if "pk" in kwargs:
            kwargs["id"] = kwargs["pk"]
            del kwargs["pk"]

        if "id" in kwargs:
            dic = self._get_from_store(int(kwargs["id"]))
        else:
            for ct in self._store.content_types.values():
                for k, v in kwargs.items():
                    if k not in ct:
                        raise ContentType.DoesNotExist()

                    if ct[k] != v:
                        break
                else:
                    dic = ct
                    break
            else:
                raise ContentType.DoesNotExist()

        def disable_save(*args, **kwargs):
            raise NotImplementedError("You can't save simulated content types")

        # We do this because some tests to comparisons with 'is' so we store
        # constructed ContentTypes in the thread local and return them if possible
        if dic["id"] in self._store.constructed_instances:
            return self._store.constructed_instances[dic["id"]]
        else:
            ContentType = self._get_model()
            result = ContentType(**dic)
            result.save = new.instancemethod(disable_save, ContentType, result)
            self._store.constructed_instances[dic["id"]] = result
            return result

    def create(self, **kwargs):
        self._repopulate_if_necessary()
        logger.warning(
            "Created simulated content type, this will not persist and will remain only on this "
            "app instance"
        )
        new_id = self._get_id(kwargs["app_label"], kwargs["model"])
        kwargs["id"] = new_id
        if "pk" in kwargs:
            del kwargs["pk"]
        self._store.content_types[new_id] = kwargs
        return self.get(id=new_id)

    def get_or_create(self, **kwargs):
        ContentType = self._get_model()
        defaults = kwargs.pop("defaults", None)
        try:
            return self.get(**kwargs), False
        except ContentType.DoesNotExist:
            if defaults:
                kwargs.update(**defaults)
            return self.create(**kwargs), True

    def filter(self, *args, **kwargs):
        self._repopulate_if_necessary()

        def _condition(ct):
            for qobj in args:
                if qobj.connector == 'AND' and not qobj.negated:
                    # normal kwargs are an AND anyway, so just use those for now
                    for child in qobj.children:
                        kwargs.update(dict([child]))
                else:
                    raise NotImplementedError("Unsupported Q object")

            for attr, val in kwargs.items():
                if getattr(ct, attr) != val:
                    return False
            return True

        # We wrap the results in a class which has a .get attribute to satisfy
        # the code which is called when a reverse relation is followed on a ForeignKey
        return ListQS([ct for ct in self.all() if _condition(ct)])

    def all(self, **kwargs):
        return self

    def __iter__(self):
        self._repopulate_if_necessary()

        for ct in self._store.content_types.keys():
            yield self.get(id=ct)

    def __len__(self):
        self._repopulate_if_necessary()
        return len(self._store.content_types)

    def using(self, *args, **kwargs):
        return self

    def bulk_create(self, *args, **kwargs):
        pass



# SYSTEM
from __future__ import absolute_import

# LIBRARIES
from django.db import models
from django.test import TestCase
from django.contrib.contenttypes.models import ContentType
from djangae.contrib.contenttypes.models import SimulatedContentTypeManager


class DummyModel(models.Model):
    pass


class RelatedToContentType(models.Model):
    content_type = models.ForeignKey(ContentType)


class SimulatedContentTypesTests(TestCase):

    def test_contenttypes_patch_is_applied(self):
        self.assertEqual(ContentType.objects.__class__, SimulatedContentTypeManager)

    def test_passing_model_to_simulated_manager_work(self):
        manager = SimulatedContentTypeManager(model=DummyModel)
        self.assertEqual(manager._get_model(), DummyModel)

    def test_get_all_contenttype_objects(self):
        self.assertTrue(len(ContentType.objects.all()) > 0)

    def test_get_for_model(self):
        ct = ContentType.objects.get_for_model(DummyModel)
        self.assertEqual(ct.model, DummyModel._meta.model_name)
        self.assertEqual(ct.app_label, DummyModel._meta.app_label)

    def test_get_for_model_not_concrete(self):
        ct = ContentType.objects.get_for_model(DummyModel, for_concrete_model=False)
        self.assertEqual(ct.model, DummyModel._meta.model_name)
        self.assertEqual(ct.app_label, DummyModel._meta.app_label)

    def test_get_by_natural_key(self):
        ct = ContentType.objects.get_by_natural_key(
            DummyModel._meta.app_label, DummyModel._meta.model_name)
        self.assertEqual(ct.model, DummyModel._meta.model_name)
        self.assertEqual(ct.app_label, DummyModel._meta.app_label)

    def test_get_for_id(self):
        ct = ContentType.objects.get_for_model(DummyModel)
        by_id = ContentType.objects.get_for_id(ct.id)
        self.assertEqual(by_id.model, DummyModel._meta.model_name)
        self.assertEqual(by_id.app_label, DummyModel._meta.app_label)

    def test_negative_ids_work(self):
        """ Test that having ContentType objects with negative IDs does not cause problems. """
        # This app_label, model combination happens to create a negative ID, which we (ab)use here
        ct, created = ContentType.objects.get_or_create(app_label='djangae', model='computedfieldmodel')
        self.assertTrue(ct.id < 0) # Make sure it's a negative ID, or this test is pointless

        related = RelatedToContentType.objects.create(content_type=ct)
        self.assertTrue(related.content_type_id < 0)

        reloaded = related.content_type
        self.assertEqual(ct.id, reloaded.id)

    def test_large_content_type_ids(self):
        """ Test that having ContentType objects with large (signed 64-bit int) IDs does not
            cause problems when other models have foreign keys to those ContentTypes.
            (Our SimulatedContentTypeManager creates signed 64-bit IDs.)
        """
        # IDs greater than > 32 bit need to work (which means patching the AutoField in ContentType to BigAutoField)
        ct = ContentType.objects.create(app_label='test', model='MyModel')
        related = RelatedToContentType.objects.create(content_type=ct)
        # Refetch from the DB to clear FK cache attributes:
        related = RelatedToContentType.objects.filter(content_type=ct).first()
        self.assertEqual(related.content_type_id, ct.pk)
        self.assertEqual(related.content_type, ct)

    def test_create_contenttype(self):
        ct = ContentType.objects.create(app_label='test', model='test')
        self.assertEqual(ct.app_label, 'test')
        self.assertEqual(ct.model, 'test')
        self.assertIsNotNone(ct.pk)

    def test_get_or_create_contenttype(self):
        ct, created = ContentType.objects.get_or_create(
            app_label=DummyModel._meta.app_label,
            model=DummyModel._meta.model_name
        )

        self.assertEqual(ct.model, DummyModel._meta.model_name)
        self.assertEqual(ct.app_label, DummyModel._meta.app_label)
        self.assertFalse(created)

        ct, created = ContentType.objects.get_or_create(
            app_label=DummyModel._meta.app_label,
            model='different_model'
        )

        self.assertEqual(ct.model, 'different_model')
        self.assertEqual(ct.app_label, DummyModel._meta.app_label)
        self.assertTrue(created)

    def test_filter_contenttypes(self):
        original_count = len(ContentType.objects.all())
        self.assertTrue(original_count > len(ContentType.objects.filter(app_label=DummyModel._meta.app_label)))



# STANDARD LIB
import logging

# THIRD PARTY
from google.appengine.ext.deferred import defer
from google.appengine.runtime import DeadlineExceededError


logger = logging.getLogger(__name__)


class Redefer(Exception):
    pass


def _process_shard(model, instance_ids, callback):
    logger.debug(
        "Processing shard for model %s, PKs %s to %s",
        model.__name__, instance_ids[0], instance_ids[-1]
    )
    for instance in model.objects.filter(pk__in=instance_ids).iterator():
        callback(instance)
    logger.debug("Done processing shard.")


def _shard(model, query, callback, shard_size, queue, offset=0):
    logger.debug("Sharding PKs for model %s into tasks on queue %s", model.__name__, queue)
    keys_queryset = model.objects.all()
    keys_queryset.query = query
    keys_queryset = keys_queryset.values_list("pk", flat=True)

    # Keep iterating until we are done, or until we might be hitting the task deadline
    shards_deferred = 0
    max_shards_to_defer_in_this_task = 250  # number which we think we can safely do in 10 minutes
    while True:
        try:
            ids = list(keys_queryset.all()[offset:offset + shard_size])
            if not ids:
                # We're done!
                logger.debug(
                    "Finished sharding PKs for model %s into tasks on queue %s.",
                    model.__name__, queue
                )
                return

            # Fire off the first shard
            defer(_process_shard, model, ids, callback, _queue=queue)
            shards_deferred += 1

            # Set the offset to the last pk
            offset += shard_size

            if shards_deferred >= max_shards_to_defer_in_this_task:
                logger.debug("Redeferring. Offset PK: %s", offset)
                raise Redefer()

        except (DeadlineExceededError, Redefer):
            # If we run out of time, or have done enough shards that we might be running out of
            # time, then defer this function again, continuing from the offset.
            defer(
                _shard,
                model,
                query,
                callback,
                shard_size,
                queue,
                offset=offset,
                _queue=queue
            )
            return


def defer_iteration(queryset, callback, shard_size=500, _queue="default", _target=None):
    """
        Shards background tasks to call 'callback' with each instance in queryset

        - `queryset` - The queryset to iterate
        - `callback` - A callable which accepts an instance as a parameter
        - `shard_size` - The number instances to process per shard (default 500)
        - `_queue` - The name of the queue to run the shards on

        Note, your callback must be indempotent, shards may retry and your callback
        may be called multiple times on the same instance. If you notice that your
        tasks are receiving DeadlineExceededErrors you probably need to reduce the
        shard size. The shards will work in parallel and will not be sequential.
    """
    # We immediately defer the _shard function so that we don't hold up execution
    defer(
        _shard, queryset.model, queryset.query, callback, shard_size, _queue,
        _queue=_queue, _target=_target,
    )



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _

try:
    from django.contrib.contenttypes.management import update_contenttypes as django_update_contenttypes
except ImportError:
    # Django 1.11
    from django.contrib.contenttypes.management import create_contenttypes as django_update_contenttypes

from django.db.models.signals import post_migrate

from .management import update_contenttypes
from .models import SimulatedContentTypeManager


class ContentTypesConfig(AppConfig):
    name = 'djangae.contrib.contenttypes'
    verbose_name = _("Djangae Content Types")
    label = "djangae_contenttypes"

    def ready(self):
        """ Patch the ContentTypes app so that:
            * The ContentType's manager is our SimulatedContentTypeManager.
            * The ContentType's PK field is BigIntegerField, so that ForeignKeys which point to it
              will acccept our large (signed 64 bit) IDs.
            * The update_contenttypes management function is replaced with our alternative version.
        """
        if django_update_contenttypes != update_contenttypes:
            post_migrate.disconnect(django_update_contenttypes)
            from django.db import models
            from django.contrib.contenttypes import models as django_models
            if not isinstance(django_models.ContentType.objects, SimulatedContentTypeManager):
                django_models.ContentType.objects = SimulatedContentTypeManager(django_models.ContentType)
                django_models.ContentType.objects.auto_created = True

                # Really force the default manager to use the Simulated one
                meta = django_models.ContentType._meta
                if hasattr(meta, "local_managers"):
                    # Django >= 1.10
                    meta.local_managers[0] = SimulatedContentTypeManager()
                else:
                    django_models.ContentType._default_manager = SimulatedContentTypeManager(django_models.ContentType)

                meta._expire_cache()

                # Our generated IDs take up a 64 bit range (signed) but aren't auto
                # incrementing so update the field to reflect that (for validation)
                meta.pk.__class__ = models.BigIntegerField



# THIRD PARTY
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import (
    AbstractBaseUser,
    python_2_unicode_compatible,
    UserManager,
)
from django.core.exceptions import ValidationError
from django.core.mail import send_mail
from django.utils.http import urlquote
from django.db import models
from django.utils import timezone, six
from django.utils.translation import ugettext_lazy as _

# DJANGAE
from djangae.fields import CharOrNoneField, ComputedCharField
from .validators import validate_google_user_id


class GaeUserManager(UserManager):

    def pre_create_google_user(self, email, **extra_fields):
        """ Pre-create a User object for a user who will later log in via Google Accounts. """
        values = dict(
            # defaults which can be overridden
            is_active=True,
        )
        values.update(**extra_fields)
        values.update(
            # things which cannot be overridden
            email=self.normalize_email(email),  # lowercase the domain only
            username=None,
            password=make_password(None),  # unusable password
            # Stupidly, last_login is not nullable, so we can't set it to None.
        )
        return self.create(**values)


def _get_email_lower(user):
    """ Computer function for the computed lowercase email field. """
    # Note that the `email` field is not nullable, but the `email_lower` one is nullable and must
    # not contain empty strings because it is unique
    return user.email and user.email.lower() or None


@python_2_unicode_compatible
class GaeAbstractBaseUser(AbstractBaseUser):
    """ Absract base class for creating a User model which works with the App
    Engine users API. """

    username = CharOrNoneField(
        # This stores the Google user_id, or custom username for non-Google-based users.
        # We allow it to be null so that Google-based users can be pre-created before they log in.
        _('User ID'), max_length=21, unique=True, blank=True, null=True, default=None,
        validators=[validate_google_user_id]
    )
    first_name = models.CharField(_('first name'), max_length=30, blank=True)
    last_name = models.CharField(_('last name'), max_length=30, blank=True)

    # Email addresses are case sensitive, but many email systems and many people treat them as if
    # they're not.  We must store the case-preserved email address to ensure that sending emails
    # always works, but we must be able to query for them case insensitively and therefore we must
    # enforce uniqueness on a case insensitive basis, hence these 2 fields
    email = models.EmailField(_('email address'))
    # The null-able-ness of the email_lower is only to deal with when an email address moves between
    # Google Accounts and therefore we need to wipe it without breaking the unique constraint.
    email_lower = ComputedCharField(
        _get_email_lower, max_length=email.max_length, unique=True, null=True
    )

    is_staff = models.BooleanField(
        _('staff status'), default=False,
        help_text=_('Designates whether the user can log into this admin site.')
    )
    is_active = models.BooleanField(
        _('active'), default=True,
        help_text=_(
            'Designates whether this user should be treated as '
            'active. Unselect this instead of deleting accounts.'
        )
    )
    date_joined = models.DateTimeField(_('date joined'), default=timezone.now)

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['email']

    objects = GaeUserManager()

    class Meta:
        abstract = True

    def clean(self):
        # Only call up if username is not none. Parent's clean() stringifies
        # username blindly
        if self.get_username() is not None:
            super(GaeAbstractBaseUser, self).clean()

    def get_absolute_url(self):
        return "/users/%s/" % urlquote(self.username)

    def get_full_name(self):
        """
        Returns the first_name plus the last_name, with a space in between.
        """
        full_name = '%s %s' % (self.first_name, self.last_name)
        return full_name.strip()

    def get_short_name(self):
        "Returns the short name for the user."
        return self.first_name

    def email_user(self, subject, message, from_email=None):
        """
        Sends an email to this User.
        """
        send_mail(subject, message, from_email, [self.email])

    def __str__(self):
        """
            We have to override this as username is nullable. We either return the email
            address, or if there is a username, we return "email (username)".
        """
        username = self.get_username()
        if username:
            return "{} ({})".format(six.text_type(self.email), six.text_type(username))
        return six.text_type(self.email)

    def validate_unique(self, exclude=None):
        """ Check that the email address does not already exist by querying on email_lower. """
        exclude = exclude or []
        if "email_lower" not in exclude:
            # We do our own check using the email_lower field, so we don't need Django to query
            # on it as well
            exclude.append("email_lower")

        try:
            super(GaeAbstractBaseUser, self).validate_unique(exclude=exclude)
        except ValidationError as super_error:
            pass
        else:
            super_error = None

        if self.email and "email" not in exclude:
            existing = self.__class__.objects.filter(email_lower=self.email.lower())
            if not self._state.adding:
                existing = existing.exclude(pk=self.pk)
            if existing.exists():
                model_name = self._meta.verbose_name
                field_name = self._meta.get_field("email").verbose_name
                message = "%s with this %s already exists" % (model_name, field_name)
                error_dict = {"email": [message]}
                if super_error:
                    super_error.update_error_dict(error_dict)
                    raise super_error
                else:
                    raise ValidationError(error_dict)
        elif super_error:
            raise



from django.contrib import auth
from django.contrib.auth.models import (
    python_2_unicode_compatible,
    GroupManager,
    _user_get_all_permissions,
    _user_has_perm,
    _user_has_module_perms
)
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.functional import lazy

from djangae.fields import ListField, RelatedSetField
from djangae.contrib.gauth.models import GaeAbstractBaseUser
from djangae.contrib.gauth_datastore.permissions import get_permission_choices


@python_2_unicode_compatible
class Group(models.Model):
    """
        This is a clone of django.contrib.auth.Group, but nonrelationalized. Doesn't user Permission but directly
        uses the permission names
    """
    name = models.CharField(_('name'), max_length=80, unique=True)
    permissions = ListField(models.CharField(max_length=500),
        verbose_name=_('permissions'), blank=True
    )

    objects = GroupManager()

    class Meta:
        db_table = 'djangae_group'
        verbose_name = _('group')
        verbose_name_plural = _('groups')

    def __str__(self):
        return self.name

    def natural_key(self):
        return (self.name,)

    def __init__(self, *args, **kwargs):
        """We need to override this to make the choices lazy and prevent import madness"""
        super(Group, self).__init__(*args, **kwargs)

        field = self._meta.get_field('permissions')
        field._choices = lazy(get_permission_choices, list)()
        field.item_field_type._choices = lazy(get_permission_choices, list)()


class PermissionsMixin(models.Model):
    """
    A mixin class that adds the fields and methods necessary to support
    Django's Group and Permission model using the ModelBackend.
    """
    is_superuser = models.BooleanField(_('superuser status'), default=False,
        help_text=_('Designates that this user has all permissions without '
                    'explicitly assigning them.')
    )

    groups = RelatedSetField(
        Group,
        verbose_name=_('groups'),
        blank=True, help_text=_('The groups this user belongs to. A user will '
                                'get all permissions granted to each of '
                                'his/her group.')
    )

    user_permissions = ListField(
        models.CharField(max_length=500),
        verbose_name=_('user permissions'), blank=True,
        help_text='Specific permissions for this user.'
    )

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        """We need to override this to make the choices lazy and prevent import madness"""
        super(PermissionsMixin, self).__init__(*args, **kwargs)
        self._meta.get_field('user_permissions')._choices = lazy(get_permission_choices, list)()

    def get_group_permissions(self, obj=None):
        """
        Returns a list of permission strings that this user has through his/her
        groups. This method queries all available auth backends. If an object
        is passed in, only permissions matching this object are returned.
        """
        permissions = set()
        for backend in auth.get_backends():
            if hasattr(backend, "get_group_permissions"):
                if obj is not None:
                    permissions.update(backend.get_group_permissions(self,
                                                                     obj))
                else:
                    permissions.update(backend.get_group_permissions(self))
        return permissions

    def get_all_permissions(self, obj=None):
        return _user_get_all_permissions(self, obj)

    def has_perm(self, perm, obj=None):
        """
        Returns True if the user has the specified permission. This method
        queries all available auth backends, but returns immediately if any
        backend returns True. Thus, a user who has permission from a single
        auth backend is assumed to have permission in general. If an object is
        provided, permissions for this specific object are checked.
        """

        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        # Otherwise we need to check the backends.
        return _user_has_perm(self, perm, obj)

    def has_perms(self, perm_list, obj=None):
        """
        Returns True if the user has each of the specified permissions. If
        object is passed, it checks if the user has all required perms for this
        object.
        """
        for perm in perm_list:
            if not self.has_perm(perm, obj):
                return False
        return True

    def has_module_perms(self, app_label):
        """
        Returns True if the user has any permissions in the given app label.
        Uses pretty much the same logic as has_perm, above.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        return _user_has_module_perms(self, app_label)


class GaeAbstractDatastoreUser(GaeAbstractBaseUser, PermissionsMixin):
    """ Base class for a user model which can be used with GAE authentication
        and permissions on the Datastore.
    """

    class Meta:
        abstract = True


class GaeDatastoreUser(GaeAbstractBaseUser, PermissionsMixin):
    """ A basic user model which can be used with GAE authentication and allows
        permissions to work on the Datastore backend.
    """

    class Meta:
        db_table = 'djangae_gaedatastoreuser'
        swappable = 'AUTH_USER_MODEL'
        verbose_name = _('user')
        verbose_name_plural = _('users')




from django import VERSION as django_version
from django.contrib.auth import authenticate, login, logout, get_user, BACKEND_SESSION_KEY, load_backend
from django.contrib.auth.middleware import AuthenticationMiddleware as DjangoMiddleware
from django.contrib.auth.models import AnonymousUser, BaseUserManager
from djangae.contrib.gauth.backends import BaseAppEngineUserAPIBackend

from google.appengine.api import users


def user_is_authenticated(django_user):
    """ Since 1.10 Django is_authenticated is an attribute, not a method.
    The backward-compatibility is supported by the function. We could remove
    the function and replace all the calls to it with
    `django_user.is_authenticated` once we stop supporting Django versions
    lower than 1.10. """
    if django_version[:3] < (1, 10, 0):
        return django_user.is_authenticated()
    return django_user.is_authenticated


class AuthenticationMiddleware(DjangoMiddleware):
    def process_request(self, request):
        django_user = get_user(request)
        google_user = users.get_current_user()

        # Check to see if the user is authenticated with a different backend, if so, just set
        # request.user and bail
        if user_is_authenticated(django_user):
            backend_str = request.session.get(BACKEND_SESSION_KEY)
            if (not backend_str) or not isinstance(load_backend(backend_str), BaseAppEngineUserAPIBackend):
                request.user = django_user
                return

        if django_user.is_anonymous() and google_user:
            # If there is a google user, but we are anonymous, log in!
            # Note that if DJANGAE_CREATE_UNKNOWN_USER=False then this may not authenticate
            django_user = authenticate(google_user=google_user) or AnonymousUser()
            if user_is_authenticated(django_user):
                login(request, django_user)

        if user_is_authenticated(django_user):
            if not google_user:
                # If we are logged in with django, but not longer logged in with Google
                # then log out
                logout(request)
                django_user = AnonymousUser()
            elif django_user.username != google_user.user_id():
                # If the Google user changed, we need to log in with the new one
                logout(request)
                django_user = authenticate(google_user=google_user) or AnonymousUser()
                if user_is_authenticated(django_user):
                    login(request, django_user)

        # Note that the logic above may have logged us out, hence new `if` statement
        if user_is_authenticated(django_user):
            self.sync_user_data(django_user, google_user)

        request.user = django_user

    def sync_user_data(self, django_user, google_user):
        # Now make sure we update is_superuser and is_staff appropriately
        changed_fields = []

        is_superuser = users.is_current_user_admin()

        if is_superuser != django_user.is_superuser:
            django_user.is_superuser = django_user.is_staff = is_superuser
            changed_fields += ['is_superuser', 'is_staff']

        email = BaseUserManager.normalize_email(google_user.email())  # Normalizes the domain only.

        if email != django_user.email:
            django_user.email = email
            changed_fields += ['email', 'email_lower']

        if changed_fields:
            django_user.save(update_fields=changed_fields)



# STANDARD LIB
from email.MIMEBase import MIMEBase
import logging

# THIRD PARTY
from django.conf import settings
from django.core.mail.backends.base import BaseEmailBackend
from django.core.mail import EmailMultiAlternatives
from google.appengine.api import mail as aeemail
from google.appengine.ext import deferred
from google.appengine.api.mail_errors import InvalidSenderError
from google.appengine.runtime import apiproxy_errors


logger = logging.getLogger(__name__)


class EmailBackend(BaseEmailBackend):
    can_defer = False

    def send_messages(self, email_messages):
        num_sent = 0
        for message in email_messages:
            if self._send(message):
                num_sent += 1
        return num_sent

    def _copy_message(self, message):
        """
        Creates and returns App Engine EmailMessage class from message.
        """
        gmsg = aeemail.EmailMessage(sender=message.from_email,
                                    to=message.to,
                                    subject=message.subject,
                                    body=message.body)
        if message.extra_headers.get('Reply-To', None):
            gmsg.reply_to = message.extra_headers['Reply-To']
        if message.cc:
            gmsg.cc = list(message.cc)
        if message.bcc:
            gmsg.bcc = list(message.bcc)
        if message.attachments:
            # Must be populated with (filename, filecontents) tuples.
            attachments = []
            for attachment in message.attachments:
                if isinstance(attachment, MIMEBase):
                    attachments.append((attachment.get_filename(),
                                        attachment.get_payload(decode=True)))
                else:
                    attachments.append((attachment[0], attachment[1]))
            gmsg.attachments = attachments
        # Look for HTML alternative content.
        if isinstance(message, EmailMultiAlternatives):
            for content, mimetype in message.alternatives:
                if mimetype == 'text/html':
                    gmsg.html = content
                    break
        return gmsg

    def _send(self, message):
        try:
            message = self._copy_message(message)
        except (ValueError, aeemail.InvalidEmailError), err:
            logger.warn(err)
            if not self.fail_silently:
                raise
            return False
        if self.can_defer:
            self._defer_send(message)
            return True
        return self._do_send(message)

    def _do_send(self, message):
        try:
            message.send()
        except (aeemail.Error, apiproxy_errors.Error) as e:
            if isinstance(e, InvalidSenderError):
                logger.error("Invalid 'from' address: %s", message.sender)
            if not self.fail_silently:
                raise
            return False
        return True

    def _defer_send(self, message):
        queue_name = getattr(settings, 'EMAIL_QUEUE_NAME', 'default')
        deferred.defer(self._do_send, message, _queue=queue_name)


class AsyncEmailBackend(EmailBackend):
    can_defer = True



# -*- coding: utf-8 -*-
# Generated by Django 1.9.9.dev20170116170623 on 2017-01-16 17:11
from __future__ import unicode_literals

from django.db import migrations, models


class AlterFieldInOtherApp(migrations.AlterField):
    """ Allows AlterField to be run on a field in a different app to the one in which the migration
        is defined by accepting an additional 'app_label' kwarg to the __init__ method.
    """

    def __init__(self, app_label, model_name, name, field, preserve_default=True):
        self.app_label = app_label
        self.model_name = model_name
        self.name = name
        self.field = field
        self.preserve_default = preserve_default

    def deconstruct(self):
        kwargs = {
            'app_label': self.app_label,
            'model_name': self.model_name,
            'name': self.name,
            'field': self.field,
        }
        if self.preserve_default is not True:
            kwargs['preserve_default'] = self.preserve_default
        return (
            self.__class__.__name__,
            [],
            kwargs
        )

    def state_forwards(self, app_label, state):
        app_label = self.app_label
        return super(AlterFieldInOtherApp, self).state_forwards(app_label, state)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        app_label = self.app_label
        return super(AlterFieldInOtherApp, self).database_forwards(app_label, schema_editor, from_state, to_state)

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        app_label = self.app_label
        return super(AlterFieldInOtherApp, self).database_backwards(app_label, schema_editor, from_state, to_state)

    def describe(self):
        return "Alter field %s on %s in app " % (self.name, self.model_name, self.app_label)


class Migration(migrations.Migration):
    """ Migration that changes the `id` field in the DJANGO ContentType app, so that foreign keys
        which point to it will allow 64 bit ints.  This then allows those foreign keys to work
        with the IDs returned by our SimulatedContentTypeManager.
    """

    dependencies = [
        ('djangae_contenttypes', '0001_patch_contenttypes_migrations'),  # DJANGAE contenttypes
        ('contenttypes', '0001_initial'),  # DJANGO contenttypes
    ]

    operations = [
        AlterFieldInOtherApp(
            app_label='contenttypes',
            model_name='contenttype',
            name='id',
            field=models.BigIntegerField(auto_created=True, blank=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



from django.conf import settings
from django.db import router
from mapreduce.mapper_pipeline import MapperPipeline
from mapreduce import parameters
from mapreduce import control
from mapreduce import model
from pipeline.util import for_name


BASE_PATH = '/_ah/mapreduce'
PIPELINE_BASE_PATH = BASE_PATH + '/pipeline'


class DjangaeMapperPipeline(MapperPipeline):

    def run(self, job_name, handler_spec, input_reader_spec, output_writer_spec=None, params=None, shards=None):
        """
            Overwriting this method allows us to pass the base_path properly, I know it's stupid but I think
            this is the cleanest way that still gives us a working Pipeline that we can chain
        """
        if shards is None:
            shards = parameters.config.SHARD_COUNT

        mapreduce_id = control.start_map(
            job_name,
            handler_spec,
            input_reader_spec,
            params or {},
            mapreduce_parameters={
                "done_callback": self.get_callback_url(),
                "done_callback_method": "GET",
                "pipeline_id": self.pipeline_id,
                "base_path": BASE_PATH,
            },
            shard_count=shards,
            output_writer_spec=output_writer_spec,
            queue_name=self.queue_name,
            )
        self.fill(self.outputs.job_id, mapreduce_id)
        self.set_status(console_url="%s/detail?mapreduce_id=%s" % (
            (BASE_PATH, mapreduce_id)))

    def callback(self):
        """
            Callback finish exists on the pipeline class, so we just use it as a nice
            wrapper for the static method attached to the MapReduceTask
        """
        super(DjangaeMapperPipeline, self).callback()

        mapreduce_id = self.outputs.job_id.value
        mapreduce_state = model.MapreduceState.get_by_job_id(mapreduce_id)
        params = mapreduce_state.mapreduce_spec.mapper.params
        finish_func = params.get('_finish', None)
        if not finish_func:
            return None
        finish_func = for_name(finish_func)
        return finish_func(*params.get('args',[]), **params.get('kwargs', {}))


class MapReduceTask(object):
    """
        MapReduceTask base class, inherit this in a statically defined class and
        use .start() to run a mapreduce task

        You must define a staticmethod 'map' which takes in an arg of the entity being mapped over.
        Optionally define a staticmethod 'reduce' for the reduce stage (Not Implemented).

        You can pass any additional args and/or kwargs to .start(), which will then be passed into
        each call of .map() for you.

        Overwrite 'finish' with a static definition for a finish callback
    """
    shard_count = None
    pipeline_class = MapperPipeline # Defaults to MapperPipeline which just runs map stage
    job_name = None
    queue_name = 'default'
    output_writer_spec = None
    mapreduce_parameters = {}
    countdown = None
    eta = None
    model = None
    map_args = []
    map_kwargs = {}
    target = 'default'


    def __init__(self, model=None, db=None):
        if model:
            self.model = model
            self.db = db or router.db_for_read(self.model)
        else:
            self.db = db or 'default'

        if not self.job_name:
            # No job name then we will just use the class
            self.job_name = self.get_class_path()

    def get_model_app_(self):
        app = self.model._meta.app_label
        name = self.model.__name__
        return '{app}.{name}'.format(
            app=app,
            name=name,
        )

    def get_class_path(self):
        return '{mod}.{cls}'.format(
            mod=self.__class__.__module__,
            cls=self.__class__.__name__,
        )

    @classmethod
    def get_relative_path(cls, func):
        return '{mod}.{cls}.{func}'.format(
            mod=cls.__module__,
            cls=cls.__name__,
            func=func.__name__,
        )

    @staticmethod
    def map(entity, *args, **kwargs):
        """
            Override this definition with a staticmethod map definition
        """
        raise NotImplementedError('You must supply a map function')

    @staticmethod
    def finish(**kwargs):
        """
            Override this with a static method for the finish callback
        """
        raise NotImplementedError('You must supply a finish function')

    def start(self, *args, **kwargs):
        # Allow a queue to be specified either as a keyword argument when
        # starting the task, or as a class-level attribute on the task.
        queue_name = kwargs.pop('queue_name', self.queue_name)

        mapper_parameters = {
            'model': self.get_model_app_(),
            'kwargs': kwargs,
            'args': args,
            'namespace': settings.DATABASES.get(self.db, {}).get("NAMESPACE"),
            'db': self.db
        }
        if 'map' not in self.__class__.__dict__:
            raise Exception('No static map method defined on class {cls}'.format(self.__class__))
        mapper_parameters['_map'] = self.get_relative_path(self.map)
        if 'finish' in self.__class__.__dict__:
            mapper_parameters['_finish'] = self.get_relative_path(self.finish)

        # Calculate a sensible shard count if one isn't specified
        shard_count = self.shard_count or (self.model.objects.using(self.db).all()[:1000].count() / 100) + 1

        pipe = DjangaeMapperPipeline(
            self.job_name,
            'djangae.contrib.mappers.thunks.thunk_map',
            'djangae.contrib.mappers.readers.DjangoInputReader',
            params=mapper_parameters,
            shards=shard_count
        )
        pipe.with_params(target=self.target)
        pipe.start(base_path=PIPELINE_BASE_PATH, queue_name=queue_name)



#STANDARD LIB
import datetime
import decimal
import warnings
import logging

#LIBRARIES
from django.conf import settings
from django.utils import timezone

from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.backends.base.client import BaseDatabaseClient
from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.backends.base.introspection import TableInfo
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.features import BaseDatabaseFeatures
from django.db.backends.base.validation import BaseDatabaseValidation
from django.db.backends.base.creation import BaseDatabaseCreation
from django.db.backends.base.schema import BaseDatabaseSchemaEditor

from google.appengine.api.datastore_types import Blob, Text
from google.appengine.datastore import datastore_stub_util
from google.appengine.api import datastore, datastore_errors

#DJANGAE
from djangae.db.utils import (
    decimal_to_string,
    make_timezone_naive,
    get_datastore_key,
)

from djangae.db.backends.appengine.caching import get_context
from djangae.db.backends.appengine.indexing import load_special_indexes
from .commands import (
    SelectCommand,
    InsertCommand,
    FlushCommand,
    UpdateCommand,
    DeleteCommand,
    coerce_unicode
)

from djangae.db.backends.appengine import dbapi as Database


logger = logging.getLogger(__name__)


class Connection(object):
    """ Dummy connection class """
    def __init__(self, wrapper, params):
        self.creation = wrapper.creation
        self.ops = wrapper.ops
        self.params = params
        self.queries = []

    def rollback(self):
        pass

    def commit(self):
        pass

    def close(self):
        pass


class Cursor(object):
    """ Dummy cursor class """
    def __init__(self, connection):
        self.connection = connection
        self.start_cursor = None
        self.returned_ids = []
        self.rowcount = -1
        self.last_select_command = None
        self.last_delete_command = None

    def execute(self, sql, *params):
        if isinstance(sql, SelectCommand):
            # Also catches subclasses of SelectCommand (e.g Update)
            self.last_select_command = sql
            self.rowcount = self.last_select_command.execute() or -1
        elif isinstance(sql, FlushCommand):
            sql.execute()
        elif isinstance(sql, UpdateCommand):
            self.rowcount = sql.execute()
        elif isinstance(sql, DeleteCommand):
            self.rowcount = sql.execute()
        elif isinstance(sql, InsertCommand):
            self.connection.queries.append(sql)
            self.returned_ids = sql.execute()
        else:
            raise Database.CouldBeSupportedError("Can't execute traditional SQL: '%s' (although perhaps we could make GQL work)", sql)

    def next(self):
        row = self.fetchone()
        if row is None:
            raise StopIteration
        return row

    def fetchone(self, delete_flag=False):
        try:
            result = self.last_select_command.results.next()

            if isinstance(result, (int, long)):
                return (result,)

            query = self.last_select_command.query

            row = []

            # Prepend extra select values to the resulting row
            for col, select in query.extra_selects:
                row.append(result.get(col))

            for col in self.last_select_command.query.init_list:
                row.append(result.get(col))

            self.returned_ids.append(result.key().id_or_name())
            return row
        except StopIteration:
            return None

    def fetchmany(self, size, delete_flag=False):
        if not self.last_select_command.results:
            return []

        result = []
        for i in xrange(size):
            entity = self.fetchone(delete_flag)
            if entity is None:
                break

            result.append(entity)

        return result

    @property
    def lastrowid(self):
        return self.returned_ids[-1].id_or_name()

    def __iter__(self):
        return self

    def close(self):
        pass

MAXINT = 9223372036854775808


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = "djangae.db.backends.appengine.compiler"

    # Datastore will store all integers as 64bit long values
    integer_field_ranges = {
        'SmallIntegerField': (-MAXINT, MAXINT-1),
        'IntegerField': (-MAXINT, MAXINT-1),
        'BigIntegerField': (-MAXINT, MAXINT-1),
        'PositiveSmallIntegerField': (0, MAXINT-1),
        'PositiveIntegerField': (0, MAXINT-1),
    }

    def bulk_batch_size(self, field, objs):
        # This value is used in cascade deletions, and also on bulk insertions
        # Bulk insertions really need to be limited to 25 elsewhere (so that they can be done)
        # transactionally, so setting to 30 doesn't matter but for cascade deletions
        # (which explode to thing_id__in=[]) we need to limit to MAX_ALLOWABLE_QUERIES
        return datastore.MAX_ALLOWABLE_QUERIES

    def quote_name(self, name):
        return name

    def date_trunc_sql(self, lookup_type, field_name):
        return ''

    def datetime_trunc_sql(self, lookup_type, field_name, tzname):
        return '', []

    def datetime_extract_sql(self, lookup_name, sql, tzname):
        return '', []

    def date_extract_sql(self, lookup_name, sql):
        return '', []

    def get_db_converters(self, expression):
        converters = super(DatabaseOperations, self).get_db_converters(expression)

        db_type = expression.field.db_type(self.connection)
        internal_type = expression.field.get_internal_type()

        if internal_type == 'TextField':
            converters.append(self.convert_textfield_value)
        elif internal_type == 'DateTimeField':
            converters.append(self.convert_datetime_value)
        elif internal_type == 'DateField':
            converters.append(self.convert_date_value)
        elif internal_type == 'TimeField':
            converters.append(self.convert_time_value)
        elif internal_type == 'DecimalField':
            converters.append(self.convert_decimal_value)
        elif db_type == 'list':
            converters.append(self.convert_list_value)
        elif db_type == 'set':
            converters.append(self.convert_set_value)

        return converters

    def convert_textfield_value(self, value, expression, connection, context=None):
        if isinstance(value, str):
            value = value.decode("utf-8")
        return value

    def convert_datetime_value(self, value, expression, connection, context=None):
        return self.connection.ops.value_from_db_datetime(value)

    def convert_date_value(self, value, expression, connection, context=None):
        return self.connection.ops.value_from_db_date(value)

    def convert_time_value(self, value, expression, connection, context=None):
        return self.connection.ops.value_from_db_time(value)

    def convert_decimal_value(self, value, expression, connection, context=None):
        return self.connection.ops.value_from_db_decimal(value)

    def convert_list_value(self, value, expression, connection, context=None):
        if expression.output_field.db_type(connection) != "list":
            return value

        if not value:
            value = []
        return value

    def convert_set_value(self, value, expression, connection, context=None):
        if expression.output_field.db_type(connection) != "set":
            return value

        if not value:
            value = set()
        else:
            value = set(value)
        return value

    def sql_flush(self, style, tables, seqs, allow_cascade=False):
        return [FlushCommand(table, self.connection) for table in tables]

    def prep_lookup_key(self, model, value, field):
        if isinstance(value, basestring):
            value = value[:500]
            left = value[500:]
            if left:
                warnings.warn("Truncating primary key that is over 500 characters. "
                              "THIS IS AN ERROR IN YOUR PROGRAM.",
                              RuntimeWarning)

            # This is a bit of a hack. Basically when you query an integer PK with a
            # string containing an int. SQL seems to return the row regardless of type, and as far as
            # I can tell, Django at no point tries to cast the value to an integer. So, in the
            # case where the internal type is an AutoField, we try to cast the string value
            # I would love a more generic solution... patches welcome!
            # It would be nice to see the SQL output of the lookup_int_as_str test is on SQL, if
            # the string is converted to an int, I'd love to know where!
            if field.get_internal_type() == 'AutoField':
                try:
                    value = int(value)
                except (TypeError, ValueError):
                    pass

            value = get_datastore_key(model, value)
        else:
            value = get_datastore_key(model, value)

        return value

    def prep_lookup_date(self, model, value, field):
        if isinstance(value, datetime.datetime):
            return value

        return self.adapt_datefield_value(value)

    def prep_lookup_time(self, model, value, field):
        if isinstance(value, datetime.datetime):
            return value

        return self.adapt_timefield_value(value)

    def prep_lookup_value(self, model, value, field, column=None):
        if field.primary_key and (not column or column == model._meta.pk.column):
            try:
                return self.prep_lookup_key(model, value, field)
            except datastore_errors.BadValueError:
                # A key couldn't be constructed from this value
                return None

        db_type = field.db_type(self.connection)

        if db_type == 'date':
            return self.prep_lookup_date(model, value, field)
        elif db_type == 'time':
            return self.prep_lookup_time(model, value, field)
        elif db_type in ('list', 'set'):
            if hasattr(value, "__len__") and not value:
                value = None #Convert empty lists to None
            elif hasattr(value, "__iter__"):
                # Convert sets to lists
                value = list(value)

        return value

    def value_for_db(self, value, field):
        if value is None:
            return None

        db_type = field.db_type(self.connection)

        if db_type in ('integer', 'long'):
            if isinstance(value, float):
                # round() always returns a float, which has a smaller max value than an int
                # so only round() it if it's already a float
                value = round(value)
            value = long(value)
        elif db_type == 'float':
            value = float(value)
        elif db_type == 'string' or db_type == 'text':
            value = coerce_unicode(value)
            if db_type == 'text':
                value = Text(value)
        elif db_type == 'bytes':
            # Store BlobField, DictField and EmbeddedModelField values as Blobs.
            value = Blob(value)
        elif db_type == 'decimal':
            value = self.adapt_decimalfield_value(value, field.max_digits, field.decimal_places)
        elif db_type in ('list', 'set'):
            if hasattr(value, "__len__") and not value:
                value = None #Convert empty lists to None
            elif hasattr(value, "__iter__"):
                # Convert sets to lists
                value = list(value)

        return value

    def last_insert_id(self, cursor, db_table, column):
        return cursor.lastrowid

    def last_executed_query(self, cursor, sql, params):
        """
            We shouldn't have to override this, but Django's BaseOperations.last_executed_query
            assumes does u"QUERY = %r" % (sql) which blows up if you have encode unicode characters
            in your SQL. Technically this is a bug in Django for assuming that sql is ASCII but
            it's only our backend that will ever trigger the problem
        """
        return u"QUERY = {}".format(sql)

    def fetch_returned_insert_id(self, cursor):
        return cursor.lastrowid

    def adapt_datetimefield_value(self, value):
        value = make_timezone_naive(value)
        return value

    def value_to_db_datetime(self, value):  # Django 1.8 compatibility
        return self.adapt_datetimefield_value(value)

    def adapt_datefield_value(self, value):
        if value is not None:
            value = datetime.datetime.combine(value, datetime.time())
        return value

    def value_to_db_date(self, value):  # Django 1.8 compatibility
        return self.adapt_datefield_value(value)

    def adapt_timefield_value(self, value):
        if value is not None:
            value = make_timezone_naive(value)
            value = datetime.datetime.combine(datetime.datetime.fromtimestamp(0), value)
        return value

    def value_to_db_time(self, value):  # Django 1.8 compatibility
        return self.adapt_timefield_value(value)

    def adapt_decimalfield_value(self, value, max_digits, decimal_places):
        if isinstance(value, decimal.Decimal):
            return decimal_to_string(value, max_digits, decimal_places)
        return value

    def value_to_db_decimal(self, value, max_digits, decimal_places):  # Django 1.8 compatibility
        return self.adapt_decimalfield_value(value, max_digits, decimal_places)

    # Unlike value_to_db, these are not overridden or standard Django, it's just nice to have symmetry
    def value_from_db_datetime(self, value):
        if isinstance(value, (int, long)):
            # App Engine Query's don't return datetime fields (unlike Get) I HAVE NO IDEA WHY
            value = datetime.datetime.fromtimestamp(float(value) / 1000000.0)

        if value is not None and settings.USE_TZ and timezone.is_naive(value):
            value = value.replace(tzinfo=timezone.utc)
        return value

    def value_from_db_date(self, value):
        if isinstance(value, (int, long)):
            # App Engine Query's don't return datetime fields (unlike Get) I HAVE NO IDEA WHY
            value = datetime.datetime.fromtimestamp(float(value) / 1000000.0)

        if value:
            value = value.date()
        return value

    def value_from_db_time(self, value):
        if isinstance(value, (int, long)):
            # App Engine Query's don't return datetime fields (unlike Get) I HAVE NO IDEA WHY
            value = datetime.datetime.fromtimestamp(float(value) / 1000000.0).time()

        if value is not None and settings.USE_TZ and timezone.is_naive(value):
            value = value.replace(tzinfo=timezone.utc)

        if value:
            value = value.time()
        return value

    def value_from_db_decimal(self, value):
        if value:
            value = decimal.Decimal(value)
        return value


class DatabaseClient(BaseDatabaseClient):
    pass


class DatabaseCreation(BaseDatabaseCreation):
    data_types = {
        'AutoField':                  'key',
        'RelatedAutoField':           'key',
        'ForeignKey':                 'key',
        'OneToOneField':              'key',
        'ManyToManyField':            'key',
        'BigIntegerField':            'long',
        'BooleanField':               'bool',
        'CharField':                  'string',
        'CommaSeparatedIntegerField': 'string',
        'DateField':                  'date',
        'DateTimeField':              'datetime',
        'DecimalField':               'decimal',
        'DurationField':              'long',
        'EmailField':                 'string',
        'FileField':                  'string',
        'FilePathField':              'string',
        'FloatField':                 'float',
        'ImageField':                 'string',
        'IntegerField':               'integer',
        'IPAddressField':             'string',
        'NullBooleanField':           'bool',
        'PositiveIntegerField':       'integer',
        'PositiveSmallIntegerField':  'integer',
        'SlugField':                  'string',
        'SmallIntegerField':          'integer',
        'TimeField':                  'time',
        'URLField':                   'string',
        'TextField':                  'text',
        'XMLField':                   'text',
        'BinaryField':                'bytes'
    }

    def __init__(self, *args, **kwargs):
        self.testbed = None
        super(DatabaseCreation, self).__init__(*args, **kwargs)

    def sql_create_model(self, model, *args, **kwargs):
        return [], {}

    def sql_for_pending_references(self, model, *args, **kwargs):
        return []

    def sql_indexes_for_model(self, model, *args, **kwargs):
        return []

    def _create_test_db(self, verbosity, autoclobber, *args):
        if args:
            logger.warning("'keepdb' argument is not currently supported on the AppEngine backend")

        get_context().reset()

    def _destroy_test_db(self, name, verbosity):
        get_context().reset()


class DatabaseIntrospection(BaseDatabaseIntrospection):
    @datastore.NonTransactional
    def get_table_list(self, cursor):
        namespace = self.connection.settings_dict.get("NAMESPACE")
        kinds = [kind.key().id_or_name() for kind in datastore.Query('__kind__', namespace=namespace).Run()]
        return [TableInfo(x, "t") for x in kinds]


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    def column_sql(self, model, field, include_default=False):
        return "", {}

    def create_model(self, model):
        """ Don't do anything when creating tables """
        pass

    def alter_unique_together(self, *args, **kwargs):
        pass

    def alter_field(self, from_model, from_field, to_field):
        pass

    def remove_field(self, from_model, field):
        pass

    def add_field(self, model, field):
        pass

    def alter_index_together(self, model, old_index_together, new_index_together):
        pass

    def delete_model(self, model):
        pass


class DatabaseFeatures(BaseDatabaseFeatures):
    empty_fetchmany_value = []
    supports_transactions = False  #FIXME: Make this True!
    can_return_id_from_insert = True
    supports_select_related = False
    autocommits_when_autocommit_is_off = True
    uses_savepoints = False
    allows_auto_pk_0 = False


class DatabaseWrapper(BaseDatabaseWrapper):

    data_types = DatabaseCreation.data_types # These moved in 1.8

    operators = {
        'exact': '= %s',
        'iexact': 'iexact %s',
        'contains': 'contains %s',
        'icontains': 'icontains %s',
        'regex': 'regex %s',
        'iregex': 'iregex %s',
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': 'startswith %s',
        'endswith': 'endswith %s',
        'istartswith': 'istartswith %s',
        'iendswith': 'iendswith %s',
        'isnull': 'isnull %s'
    }

    Database = Database

    # These attributes are only used by Django >= 1.11
    client_class = DatabaseClient
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    features_class = DatabaseFeatures
    ops_class = DatabaseOperations
    creation_class = DatabaseCreation
    validation_class = BaseDatabaseValidation

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)

        if not hasattr(self, "client"):
            # Django 1.11 creates these automatically, when we call super
            # These are here for Django <= 1.10
            self.features = DatabaseFeatures(self)
            self.ops = DatabaseOperations(self)
            self.client = DatabaseClient(self)
            self.creation = DatabaseCreation(self)
            self.introspection = DatabaseIntrospection(self)
            self.validation = BaseDatabaseValidation(self)

        self.autocommit = True

    def is_usable(self):
        return True

    def get_connection_params(self):
        return {}

    def get_new_connection(self, params):
        conn = Connection(self, params)
        load_special_indexes()  # make sure special indexes are loaded
        return conn

    def init_connection_state(self):
        pass

    def _start_transaction_under_autocommit(self):
        pass

    def _set_autocommit(self, enabled):
        self.autocommit = enabled

    def create_cursor(self, name=None):
        self.name = name  # Django >= 1.11
        if not self.connection:
            self.connection = self.get_new_connection(self.settings_dict)

        return Cursor(self.connection)

    def schema_editor(self, *args, **kwargs):
        return DatabaseSchemaEditor(self, *args, **kwargs)

    def validate_no_broken_transaction(self):
        # Override this to do nothing, because it's not relevant to the Datastore
        pass



from django import template

register = template.Library()

def blobstore_upload_url(destination_url):
    from google.appengine.ext.blobstore import create_upload_url
    return create_upload_url(unicode(destination_url))

def gcs_upload_url(destination_url, bucket_name):
    from google.appengine.ext.blobstore import create_upload_url
    return create_upload_url(unicode(destination_url), gs_bucket_name=unicode(bucket_name))

register.simple_tag(blobstore_upload_url)
register.simple_tag(gcs_upload_url)



import copy
import collections

from google.appengine.api import datastore

class CopyDict(collections.MutableMapping):
    """
        It's important we don't pass references around in and out
        of the cache, so this just ensures we copy stuff going in, and
        copy it going out!
    """
    def __init__(self, *args, **kwargs):
        self._store = {}
        super(CopyDict, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        value = copy.deepcopy(value)
        self._store[key] = value

    def __getitem__(self, key):
        value = self._store[key]
        return copy.deepcopy(value)

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)


class ContextCache(object):
    """ Object via which the stack of Context objects and the settings for the context caching are
        accessed. A separate instance of this should exist per thread.
    """
    def __init__(self):
        self.memcache_enabled = True
        self.context_enabled = True
        self.stack = ContextStack()

    def reset(self, keep_disabled_flags=False):
        if datastore.IsInTransaction():
            raise RuntimeError(
                "Clearing the context cache inside a transaction breaks everything, "
                "we can't let you do that"
            )
        self.stack = ContextStack()
        if not keep_disabled_flags:
            self.memcache_enabled = True
            self.context_enabled = True


class Context(object):

    def __init__(self, stack):
        self.cache = CopyDict()
        self.reverse_cache = CopyDict()
        self._stack = stack

    def apply(self, other):
        self.cache.update(other.cache)

        # We have to delete things that don't exist in the other
        for k in self.cache.keys():
            if k not in other.cache:
                del self.cache[k]

        self.reverse_cache.update(other.reverse_cache)

        # We have to delete things that don't exist in the other
        for k in self.reverse_cache.keys():
            if k not in other.reverse_cache:
                del self.reverse_cache[k]

    def cache_entity(self, identifiers, entity, situation):
        assert hasattr(identifiers, "__iter__")

        for identifier in identifiers:
            self.cache[identifier] = copy.deepcopy(entity)

        self.reverse_cache[entity.key()] = identifiers

    def remove_entity(self, entity_or_key):
        if not isinstance(entity_or_key, datastore.Key):
            entity_or_key = entity_or_key.key()

        for identifier in self.reverse_cache[entity_or_key]:
            del self.cache[identifier]

        del self.reverse_cache[entity_or_key]

    def get_entity(self, identifier):
        return self.cache.get(identifier)

    def get_entity_by_key(self, key):
        try:
            identifier = self.reverse_cache[key][0]
        except KeyError:
            return None
        return self.get_entity(identifier)


class ContextStack(object):
    """
        A stack of contexts. This is used to support in-context
        caches for multi level transactions.
    """

    def __init__(self):
        self.stack = [Context(self)]
        self.staged = []

    def push(self):
        self.stack.append(
            Context(self) # Empty context
        )

    def pop(self, apply_staged=False, clear_staged=False, discard=False):
        """
            apply_staged: pop normally takes the top of the stack and adds it to a FIFO
            queue. By passing apply_staged it will pop to the FIFO queue then apply the
            queue to the top of the stack.

            clear_staged: pop, and then wipe out any staged contexts.

            discard: Ignores the popped entry in the stack, it's just discarded

            The staged queue will be wiped out if the pop makes the size of the stack one,
            regardless of whether you pass clear_staged or not. This is for safety!
        """
        from . import caching


        if not discard:
            self.staged.insert(0, self.stack.pop())
        else:
            self.stack.pop()

        if apply_staged:
            while self.staged:
                to_apply = self.staged.pop()
                keys = to_apply.reverse_cache.keys()
                if keys:
                    # This assumes that all keys are in the same namespace, which is almost definitely
                    # going to be the case, but it feels a bit dirty
                    namespace = keys[0].namespace() or None
                    caching.remove_entities_from_cache_by_key(
                        keys, namespace=namespace, memcache_only=True
                    )

                self.top.apply(to_apply)

        if clear_staged or len(self.stack) == 1:
            self.staged = []

    @property
    def top(self):
        return self.stack[-1]

    @property
    def size(self):
        return len(self.stack)

    @property
    def staged_count(self):
        return len(self.staged)



# 3RD PARTY
from django.db import models

# DJANGAE
from djangae.contrib.consistency.consistency import (
    handle_post_save,
    handle_post_delete,
)


POST_SAVE_UID = "consistency_post_save"
POST_DELETE_UID = "consistency_post_delete"


def connect_signals():
    models.signals.post_save.connect(handle_post_save, dispatch_uid=POST_SAVE_UID)
    models.signals.post_save.connect(handle_post_delete, dispatch_uid=POST_DELETE_UID)


def disconnect_signals():
    models.signals.post_save.disconnect(dispatch_uid=POST_SAVE_UID)
    models.signals.post_save.disconnect(dispatch_uid=POST_DELETE_UID)





from djangae.environment import task_or_admin_only
from djangae.test import TestCase
from djangae.contrib import sleuth
from django.http import HttpResponse


class TaskOrAdminOnlyTestCase(TestCase):
    """ Tests for the @task_or_admin_only decorator. """

    def test_403_if_not_task_or_admin(self):
        # If we are neither in a task or logged in as an admin, we expect a 403 response

        @task_or_admin_only
        def view(request):
            return HttpResponse("Hello")

        response = view(None)
        self.assertEqual(response.status_code, 403)

    def test_allowed_if_in_task(self):
        """ If we're in an App Engine task then it should allow the request through. """

        @task_or_admin_only
        def view(request):
            return HttpResponse("Hello")

        with sleuth.fake("djangae.environment.is_in_task", True):
            response = view(None)
        self.assertEqual(response.status_code, 200)

    def test_allowed_if_in_cron(self):
        """ If the view is being called by the GAE cron, then it should allow the request through. """

        @task_or_admin_only
        def view(request):
            return HttpResponse("Hello")

        with sleuth.fake("djangae.environment.is_in_cron", True):
            response = view(None)
        self.assertEqual(response.status_code, 200)

    def test_allowed_if_admin_user(self):
        """ If we're logged in as an admin of the GAE application then we should allow through. """

        @task_or_admin_only
        def view(request):
            return HttpResponse("Hello")

        with sleuth.fake("djangae.environment.users.is_current_user_admin", True):
            response = view(None)
        self.assertEqual(response.status_code, 200)



from djangae.forms.fields import TrueOrNullFormField
from djangae.core import validators
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _
from google.appengine.api.datastore_types import _MAX_STRING_LENGTH

from .iterable import *
from .related import *
from .computed import *
from .json import *
from .counting import *


class TrueOrNullField(models.NullBooleanField):
    """A Field only storing `Null` or `True` values.

    Why? This allows unique_together constraints on fields of this type
    ensuring that only a single instance has the `True` value.

    It mimics the NullBooleanField field in it's behaviour, while it will
    raise an exception when explicitly validated, assigning something
    unexpected (like a string) and saving, will silently convert that
    value to either True or None.
    """
    default_error_messages = {
        'invalid': _("'%s' value must be either True or None."),
    }
    description = _("Boolean (Either True or None)")

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_python(self, value):
        if value in (None, 'None', False):
            return None
        if value in (True, 't', 'True', '1'):
            return True
        msg = self.error_messages['invalid'] % str(value)
        raise ValidationError(msg)

    def get_prep_value(self, value):
        """Only ever save None's or True's in the db. """
        if value in (None, False, '', 0):
            return None
        return True

    def formfield(self, **kwargs):
        defaults = {
            'form_class': TrueOrNullFormField
        }
        defaults.update(kwargs)
        return super(TrueOrNullField, self).formfield(**defaults)


class CharOrNoneField(models.CharField):
    """ A field that stores only non-empty strings or None (it won't store empty strings).
        This is useful if you want values to be unique but also want to allow empty values.
    """
    empty_strings_allowed = False

    def __init__(self, *args, **kwargs):
        # Don't allow null=False because that would be insane.
        if not kwargs.get('null', True):
            raise ImproperlyConfigured("You can't set null=False on a CharOrNoneField.")
        # Set blank=True as the default, but allow it to be overridden, as it's theoretically
        # possible that you might want to prevent emptiness only in a form
        defaults = dict(null=True, blank=True, default=None)
        defaults.update(**kwargs)
        super(CharOrNoneField, self).__init__(*args, **defaults)

    def pre_save(self, model_instance, add):
        value = super(CharOrNoneField, self).pre_save(model_instance, add)
        # Change empty strings to None
        if not value:
            return None
        return value


class CharField(models.CharField):

    def __init__(self, max_length=_MAX_STRING_LENGTH, *args, **kwargs):
        assert max_length <= _MAX_STRING_LENGTH, \
            "%ss max_length must not be grater than %d bytes." % (self.__class__.__name__, _MAX_STRING_LENGTH)

        super(CharField, self).__init__(max_length=max_length, *args, **kwargs)

        # Append the MaxBytesValidator if it's not been included already
        self.validators = [
            x for x in self.validators if not isinstance(x, validators.MaxBytesValidator)
        ] + [validators.MaxBytesValidator(limit_value=max_length)]



from django.conf.urls import url, include

import djangae.contrib.mappers.urls
from . import views


urlpatterns = [
    url(r'^start$', views.start),
    url(r'^stop$', views.stop),
    url(r'^warmup$', views.warmup),
    url(r'^clearsessions$', views.clearsessions),
    url(r'^queue/deferred/?$', views.deferred),
    url(r'^internalupload/$', views.internalupload, name='djangae_internal_upload_handler'),
    url(r'^mapreduce/', include(djangae.contrib.mappers.urls)),
]



import os
import re
import sys
import argparse

import djangae.sandbox as sandbox
from djangae import environment

# Set some Django-y defaults
DJANGO_DEFAULTS = {
    "storage_path": os.path.join(environment.get_application_root(), ".storage"),
    "port": 8000,
    "admin_port": 8001,
    "api_port": 8002,
    "automatic_restart": "True",
    "allow_skipped_files": "True",
}


def _execute_from_command_line(sandbox_name, argv, **sandbox_overrides):
    # Parses for a settings flag, adding it as an environment variable to
    # the sandbox if found.
    parser = argparse.ArgumentParser(prog='manage.py')
    parser.add_argument('--settings', nargs='?')
    parsed = parser.parse_known_args(argv)
    settings = parsed[0].settings
    env_vars = {}
    if settings:
        env_vars['DJANGO_SETTINGS_MODULE'] = settings

    # retrieve additional overridden module settings
    for arg in parsed[1]:
        m = re.match(r'--(?P<module_name>.+)-settings=(?P<settings_path>.+)', arg)
        if m:
            argv.remove(arg)
            env_vars['%s_DJANGO_SETTINGS_MODULE' % m.group('module_name')] = m.group('settings_path')

    with sandbox.activate(
        sandbox_name,
        add_sdk_to_path=True,
        new_env_vars=env_vars,
        **sandbox_overrides
    ):
        import django.core.management as django_management  # Now on the path
        return django_management.execute_from_command_line(argv)


def execute_from_command_line(argv=None, **sandbox_overrides):
    """Wraps Django's `execute_from_command_line` to initialize a djangae
    sandbox before running a management command.
    """
    argv = argv or sys.argv
    parser = argparse.ArgumentParser(prog='manage.py')
    parser.add_argument(
        '--sandbox', default=sandbox.LOCAL, choices=sandbox.SANDBOXES.keys())
    parser.add_argument('--app_id', default=None, help='GAE APPLICATION ID')
    namespace, other_args = parser.parse_known_args(argv[1:])

    overrides = DJANGO_DEFAULTS
    overrides.update(sandbox_overrides, app_id=namespace.app_id)

    return _execute_from_command_line(namespace.sandbox, ['manage.py'] + other_args, **overrides)


def remote_execute_from_command_line(argv=None, **sandbox_overrides):
    """Execute commands in the remote sandbox"""
    return _execute_from_command_line(sandbox.REMOTE, argv or sys.argv, **sandbox_overrides)


def local_execute_from_command_line(argv=None, **sandbox_overrides):
    """Execute commands in the local sandbox"""
    return _execute_from_command_line(sandbox.LOCAL, argv or sys.argv, **sandbox_overrides)


def test_execute_from_command_line(argv=None, **sandbox_overrides):
    """Execute commands in the test sandbox"""
    return _execute_from_command_line(sandbox.TEST, argv or sys.argv, **sandbox_overrides)



import time
from datetime import datetime

from google.appengine.api import memcache

class MemcacheLock(object):
    def __init__(self, identifier, cache):
        self.identifier = identifier
        self._cache = cache

    @classmethod
    def acquire(cls, identifier, wait=True, steal_after_ms=None):
        cache = memcache.Client()

        def expired(value):
            if not value:
                # No value? We're fine
                return True

            return (datetime.utcnow() - value).total_seconds() * 1000 > steal_after_ms

        acquired = cache.add(identifier, datetime.utcnow())
        if acquired:
            # We set the key so we have the lock
            return cls(identifier, cache)
        elif not wait:
            return None
        else:
            while True:
                # We are waiting for a while
                to_set = datetime.utcnow()

                # Use compare-and-set to atomically set our new timestamp
                # if the lock has expired
                stored = cache.gets(identifier)
                if expired(stored):
                    if cache.cas(identifier, to_set):
                        break

                time.sleep(0.1)

            return cls(identifier, cache)

    def release(self):
        cache = self._cache
        cache.delete(self.identifier)



from django.contrib.auth.models import PermissionsMixin
from django.utils.translation import ugettext_lazy as _

from djangae.contrib.gauth.models import GaeAbstractBaseUser


class GaeAbstractUser(GaeAbstractBaseUser, PermissionsMixin):
    """
    Abstract user class for SQL databases.
    """
    class Meta:
        abstract = True


class GaeUser(GaeAbstractBaseUser, PermissionsMixin):
    """ A basic user model which can be used with GAE authentication.
        Essentially the equivalent of django.contrib.auth.models.User.
        Cannot be used with permissions when using the Datastore, because it
        uses the standard django permissions models which use M2M relationships.
    """

    class Meta:
        swappable = 'AUTH_USER_MODEL'
        verbose_name = _('user')
        verbose_name_plural = _('users')



from django import forms
from django.db import models
from django.contrib import admin
from .models import UniqueAction, ActionLog, encode_model

def _show_model(m):

    if m._meta.app_label == "uniquetool":
        return False

    if any([ x.unique for x in m._meta.fields ]):
        return True

    if x._meta.unique_together:
        return True

    return False

class ActionLogInline(admin.TabularInline):
    model = ActionLog
    verbose_name_plural = 'List of action messages'

    can_delete = False
    extra = 0
    editable_fields = []
    readonly_fields = ('log_type', 'instance_key', 'marker_key', )

    def has_add_permission(self, request):
        return False


class UniqueActionAdmin(admin.ModelAdmin):
    actions = None
    list_display = ('action_type', 'model_name', 'status')
    change_form_template = "admin/unique_action_change_form.html"
    inlines = [ActionLogInline]

    @classmethod
    def model_choices(cls):
        if not hasattr(cls, '_model_choices'):
            all_models = sorted([
                (encode_model(m), m.__name__)
                for m in models.get_models()
                if _show_model(m)
            ], key=lambda x: x[1])
            cls._model_choices = all_models
        return cls._model_choices

    def model_name(self, instance):
        return dict(self.model_choices())[instance.model]

    def get_form(self, request, obj=None, **kwargs):
        if obj is not None and obj.pk:
            kwargs['fields'] = []
            return super(UniqueActionAdmin, self).get_form(request, obj=obj, **kwargs)

        form = super(UniqueActionAdmin, self).get_form(request, obj=obj, **kwargs)
        # FIXME: this field should be optional when a "clean" action is selected
        form.base_fields['model'] = forms.ChoiceField(choices=self.model_choices())
        return form

    def has_delete_permission(self, request, obj=None):
        if obj and obj.status == 'running':
            return False
        return super(UniqueActionAdmin, self).has_delete_permission(request, obj=obj)

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        if obj:
            context["title"] = u"Errors from %s on %s (%s)" % (obj.action_type, self.model_name(obj), obj.get_status_display())
            context["readonly"] = True
        return super(UniqueActionAdmin, self).render_change_form(request, context, add=add, change=change, form_url=form_url, obj=obj)


admin.site.register(UniqueAction, UniqueActionAdmin)



# STANDARD LIB
import threading

# DJANGAE
from djangae.db import transaction
from djangae.contrib import sleuth
from djangae.test import TestCase


class TransactionTests(TestCase):
    def test_repeated_usage_in_a_loop(self):
        from .test_connector import TestUser
        pk = TestUser.objects.create(username="foo").pk
        for i in xrange(4):
            with transaction.atomic(xg=True):
                TestUser.objects.get(pk=pk)
                continue

        with transaction.atomic(xg=True):
            TestUser.objects.get(pk=pk)

    def test_recursive_atomic(self):
        l = []

        @transaction.atomic
        def txn():
            l.append(True)
            if len(l) == 3:
                return
            else:
                txn()

        txn()

    def test_recursive_non_atomic(self):
        l = []

        @transaction.non_atomic
        def txn():
            l.append(True)
            if len(l) == 3:
                return
            else:
                txn()

        txn()

    def test_atomic_in_separate_thread(self):
        """ Regression test.  See #668. """
        @transaction.atomic
        def txn():
            return

        def target():
            txn()

        thread = threading.Thread(target=target)
        thread.start()
        thread.join()

    def test_non_atomic_in_separate_thread(self):
        """ Regression test.  See #668. """
        @transaction.non_atomic
        def txn():
            return

        def target():
            txn()

        thread = threading.Thread(target=target)
        thread.start()
        thread.join()

    def test_atomic_decorator(self):
        from .test_connector import TestUser

        @transaction.atomic
        def txn():
            TestUser.objects.create(username="foo", field2="bar")
            self.assertTrue(transaction.in_atomic_block())
            raise ValueError()

        with self.assertRaises(ValueError):
            txn()

        self.assertEqual(0, TestUser.objects.count())

    def test_interaction_with_datastore_txn(self):
        from google.appengine.ext import db
        from google.appengine.datastore.datastore_rpc import TransactionOptions
        from .test_connector import TestUser

        @db.transactional(propagation=TransactionOptions.INDEPENDENT)
        def some_indie_txn(_username):
            TestUser.objects.create(username=_username)

        @db.transactional()
        def some_non_indie_txn(_username):
            TestUser.objects.create(username=_username)

        @db.transactional()
        def double_nested_transactional():
            @db.transactional(propagation=TransactionOptions.INDEPENDENT)
            def do_stuff():
                TestUser.objects.create(username="Double")
                raise ValueError()

            try:
                return do_stuff
            except:
                return

        with transaction.atomic():
            double_nested_transactional()

        @db.transactional()
        def something_containing_atomic():
            with transaction.atomic():
                TestUser.objects.create(username="Inner")

        something_containing_atomic()

        with transaction.atomic():
            with transaction.atomic():
                some_non_indie_txn("Bob1")
                some_indie_txn("Bob2")
                some_indie_txn("Bob3")

        with transaction.atomic(independent=True):
            some_non_indie_txn("Fred1")
            some_indie_txn("Fred2")
            some_indie_txn("Fred3")

    def test_atomic_context_manager(self):
        from .test_connector import TestUser

        with self.assertRaises(ValueError):
            with transaction.atomic():
                TestUser.objects.create(username="foo", field2="bar")
                raise ValueError()

        self.assertEqual(0, TestUser.objects.count())

    def test_non_atomic_context_manager(self):
        from .test_connector import TestUser
        existing = TestUser.objects.create(username="existing", field2="exists")

        with transaction.atomic():
            self.assertTrue(transaction.in_atomic_block())

            user = TestUser.objects.create(username="foo", field2="bar")

            with transaction.non_atomic():
                # We're outside the transaction, so the user should not exist
                self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user.pk)
                self.assertFalse(transaction.in_atomic_block())

                with sleuth.watch("google.appengine.api.datastore.Get") as datastore_get:
                    TestUser.objects.get(pk=existing.pk)  # Should hit the cache, not the datastore

                self.assertFalse(datastore_get.called)

            with transaction.atomic(independent=True):
                user2 = TestUser.objects.create(username="foo2", field2="bar2")
                self.assertTrue(transaction.in_atomic_block())

                with transaction.non_atomic():
                    self.assertFalse(transaction.in_atomic_block())
                    self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user2.pk)

                    with transaction.non_atomic():
                        self.assertFalse(transaction.in_atomic_block())
                        self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user2.pk)

                        with sleuth.watch("google.appengine.api.datastore.Get") as datastore_get:
                            # Should hit the cache, not the Datastore
                            TestUser.objects.get(pk=existing.pk)

                    self.assertFalse(transaction.in_atomic_block())
                    self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user2.pk)

                self.assertTrue(TestUser.objects.filter(pk=user2.pk).exists())
                self.assertTrue(transaction.in_atomic_block())

    def test_xg_argument(self):
        from .test_connector import TestUser, TestFruit

        @transaction.atomic(xg=True)
        def txn(_username):
            TestUser.objects.create(username=_username, field2="bar")
            TestFruit.objects.create(name="Apple", color="pink")
            raise ValueError()

        with self.assertRaises(ValueError):
            txn("foo")

        self.assertEqual(0, TestUser.objects.count())
        self.assertEqual(0, TestFruit.objects.count())

    def test_independent_argument(self):
        """
            We would get a XG error if the inner transaction was not independent
        """
        from .test_connector import TestUser, TestFruit

        @transaction.atomic
        def txn1(_username, _fruit):
            @transaction.atomic(independent=True)
            def txn2(_fruit):
                TestFruit.objects.create(name=_fruit, color="pink")
                raise ValueError()

            TestUser.objects.create(username=_username)
            txn2(_fruit)

        with self.assertRaises(ValueError):
            txn1("test", "banana")

    def test_nested_decorator(self):
        # Nested decorator pattern we discovered can cause a connection_stack
        # underflow.

        @transaction.atomic
        def inner_txn():
            pass

        @transaction.atomic
        def outer_txn():
            inner_txn()

        # Calling inner_txn first puts it in a state which means it doesn't
        # then behave properly in a nested transaction.
        inner_txn()
        outer_txn()



from django.contrib import admin

# DJANGAE
from djangae.contrib.gauth_datastore.models import (
    GaeDatastoreUser,
    Group
)


@admin.register(GaeDatastoreUser)
class UserAdmin(admin.ModelAdmin):
    exclude = ('password',)

    def save_model(self, request, user, form, change):
        if not user.password:
            user.set_password(None)
        user.save()

admin.site.register(Group)



from django.core.validators import BaseValidator, MinLengthValidator, MaxLengthValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import ungettext_lazy


class MaxBytesValidator(BaseValidator):
    compare = lambda self, a, b: a > b
    clean = lambda self, x: len(x.encode('utf-8'))
    message = ungettext_lazy(
        'Ensure this value has at most %(limit_value)d byte (it has %(show_value)d).',
        'Ensure this value has at most %(limit_value)d bytes (it has %(show_value)d).',
        'limit_value')
    code = 'max_length'


@deconstructible
class MinItemsValidator(MinLengthValidator):
    """ Copy of MinLengthValidator, but with the message customised to say "items" instead of
        "characters".
    """
    message = ungettext_lazy(
        'Ensure this field has at least %(limit_value)d item (it has %(show_value)d).',
        'Ensure this field has at least %(limit_value)d items (it has %(show_value)d).',
        'limit_value')


@deconstructible
class MaxItemsValidator(MaxLengthValidator):
    """ Copy of MaxLengthValidator, but with the message customised to say "items" instead of
        "characters".
    """
    message = ungettext_lazy(
        'Ensure this field has at most %(limit_value)d item (it has %(show_value)d).',
        'Ensure this field has at most %(limit_value)d items (it has %(show_value)d).',
        'limit_value')



from django.db import NotSupportedError
from django.db.models.expressions import F, Col
from djangae.db.utils import get_prepared_db_value
from django.db.models.aggregates import Aggregate

CONNECTORS = {
    F.ADD: lambda l, r: l + r,
    F.SUB: lambda l, r: l - r,
    F.MUL: lambda l, r: l * r,
    F.DIV: lambda l, r: l / r,
}


def evaluate_expression(expression, instance, connection):
    """ A limited evaluator for Django's F expressions. This are evaluated within
        the get/put transaction in _update_entity so these will happen atomically
    """

    if isinstance(expression, (basestring, int, float)):
        return expression

    if isinstance(expression, Aggregate):
        raise NotSupportedError("Aggregate expressions are not supported on the datastore")

    if hasattr(expression, 'name'):
        field = instance._meta.get_field(expression.name)
        return get_prepared_db_value(connection, instance._original, field)

    if isinstance(expression, Col):
        field = expression.field
        return get_prepared_db_value(connection, instance._original, field)

    if hasattr(expression, 'value'):
        return expression.value

    if hasattr(expression, 'connector') and expression.connector in CONNECTORS:
        if hasattr(expression, 'children'):
            lhs, rhs = expression.children
        else:
            lhs, rhs = expression.lhs, expression.rhs

        return CONNECTORS[expression.connector](
            evaluate_expression(lhs, instance, connection),
            evaluate_expression(rhs, instance, connection),
        )

    raise NotImplementedError("Support for expression %r isn't implemented", expression)



from djangae.test import TestCase
from djangae.contrib.security.management.commands import dumpurls


class DumpUrlsTests(TestCase):
    def test_dumpurls(self):
        print ("*" * 50) + "\n\n\n"
        """ Test that the `dumpurls` command runs without dying. """
        command = dumpurls.Command()
        command.handle()



# -*- coding: utf-8 -*-
import os
from importlib import import_module

from django.db import migrations
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.operations.base import Operation
from djangae.contrib.contenttypes.models import SimulatedContentTypeManager


class PatchMigrationsOperation(Operation):

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        pass

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        pass

    def state_forwards(self, app_label, state):
        """ Patch manager on ModelState during migrations """
        contenttype = state.models[('contenttypes', 'contenttype')]
        contenttype.managers[0] = ('objects', SimulatedContentTypeManager(model=contenttype))


def get_installed_app_labels_with_migrations():
    """ Get the app labels, because settings.INSTALLED_APPS doesn't necessarily give us the labels.
        Remove django.contrib.contenttypes because we want it to run before us.
        Return list of tuples like ('admin', '__first__')
    """
    from django.apps import apps
    apps_with_migrations = []
    for app in apps.get_app_configs():
        if app.label == 'contenttypes':
            continue  # Ignore the contenttypes app

        migrations_module = MigrationLoader.migrations_module(app.label)
        try:
            # Django 1.11 changed the return value of the migrations_module call to a 2-element
            # tuple.  The actual module is the first entry
            if isinstance(migrations_module, tuple):
                migrations_module = migrations_module[0]

            module = import_module(migrations_module)
        except ImportError:
            continue

        if not hasattr(module, "__path__"):
            continue

        # Make sure there are python files in the migration folder (other than the init file)
        has_files = any(
            x for x in os.listdir(module.__path__[0]) if x.endswith(".py") and x != "__init__.py"
        )
        if not has_files:
            continue

        apps_with_migrations.append(app.label)

    return [(x, '__first__') for x in apps_with_migrations]


class Migration(migrations.Migration):

    run_before = get_installed_app_labels_with_migrations()

    dependencies = [
        ('contenttypes', '__first__')
    ]

    operations = [
        PatchMigrationsOperation(),
    ]



import django
import logging
import yaml
import os
import datetime
import re

from django.db import models
from django.apps import apps
from django.conf import settings

from djangae import environment
from djangae.fields import iterable
from djangae.sandbox import allow_mode_write


logger = logging.getLogger(__name__)
_project_special_indexes = {}
_app_special_indexes = {}
_last_loaded_times = {}
_indexes_loaded = False


MAX_COLUMNS_PER_SPECIAL_INDEX = getattr(settings, "DJANGAE_MAX_COLUMNS_PER_SPECIAL_INDEX", 3)
CHARACTERS_PER_COLUMN = [31, 44, 54, 63, 71, 79, 85, 91, 97, 103]
STRIP_PERCENTS = django.VERSION < (1, 10)


def _get_project_index_file():
    project_index_file = os.path.join(environment.get_application_root(), "djangaeidx.yaml")
    return project_index_file


def _get_app_index_files():
    index_files = []

    for app_config in apps.get_app_configs():
        app_path = app_config.path
        project_index_file = os.path.join(app_path, "djangaeidx.yaml")
        index_files.append(project_index_file)
    return index_files


def _get_table_from_model(model_class):
    return model_class._meta.db_table.encode("utf-8")


def _is_iterable(value):
    return hasattr(value, '__iter__')  # is a list, tuple or set?


def _deduplicate_list(value_list):
    """ Deduplicate list of elements; value_list is expected to be a list
    of containing hashable elements. """
    return list(set(value_list))


def _make_lower(value):
    """ Make string and list of strings lowercase """
    if _is_iterable(value):
        return [v.lower() for v in value]
    else:
        return value.lower()


def _merged_indexes():
    """
        Returns the combination of the app and project special indexes
    """
    global _project_special_indexes
    global _app_special_indexes

    result = _app_special_indexes.copy()
    for model, indexes in _project_special_indexes.items():
        for field_name, values in indexes.items():
            result.setdefault(
                model, {}
            ).setdefault(field_name, []).extend(values)
    return result


def load_special_indexes():
    global _project_special_indexes
    global _app_special_indexes
    global _last_loaded_times
    global _indexes_loaded

    if _indexes_loaded and environment.is_production_environment():
        # Index files can't change if we're on production, so once they're loaded we don't need
        # to check their modified times and reload them
        return

    def _read_file(filepath):
        # Load any existing indexes
        with open(filepath, "r") as stream:
            data = yaml.load(stream)
        return data

    project_index_file = _get_project_index_file()
    app_files = _get_app_index_files()

    files_to_reload = {}

    # Go through and reload any files that we find
    for file_path in [project_index_file] + app_files:
        if not os.path.exists(file_path):
            continue

        mtime = os.path.getmtime(file_path)
        if _last_loaded_times.get(file_path) and _last_loaded_times[file_path] == mtime:
            # The file hasn't changed since last time, so do nothing
            continue
        else:
            # Mark this file for reloading, store the current modified time
            files_to_reload[file_path] = mtime

    # First, reload the project index file,
    if project_index_file in files_to_reload:
        mtime = files_to_reload[project_index_file]
        _project_special_indexes = _read_file(project_index_file)
        _last_loaded_times[project_index_file] = mtime

        # Remove it from the files to reload
        del files_to_reload[project_index_file]

    # Now, load the rest of the files and update any entries
    for file_path in files_to_reload:
        mtime = files_to_reload[project_index_file]
        new_data = _read_file(file_path)
        _last_loaded_times[file_path] = mtime

        # Update the app special indexes list
        for model, indexes in new_data.items():
            for field_name, values in indexes.items():
                _app_special_indexes.setdefault(
                    model, {}
                ).setdefault(field_name, []).extend(values)

    _indexes_loaded = True
    logger.debug("Loaded special indexes for %d models", len(_merged_indexes()))


def special_index_exists(model_class, field_name, index_type):
    table = _get_table_from_model(model_class)
    return index_type in _merged_indexes().get(table, {}).get(field_name, [])


def special_indexes_for_model(model_class):
    classes = [model_class] + model_class._meta.parents.keys()

    result = {}
    for klass in classes:
        result.update(_merged_indexes().get(_get_table_from_model(klass), {}))
    return result


def special_indexes_for_column(model_class, column):
    return special_indexes_for_model(model_class).get(column, [])


def write_special_indexes():
    """
        Writes the project-specific indexes to the project djangaeidx.yaml
    """
    project_index_file = _get_project_index_file()

    with allow_mode_write():
        with open(project_index_file, "w") as stream:
            stream.write(yaml.dump(_project_special_indexes))


def add_special_index(model_class, field_name, indexer, operator, value=None):
    from djangae.utils import in_testing
    from django.conf import settings

    index_type = indexer.prepare_index_type(operator, value)

    field_name = field_name.encode("utf-8")  # Make sure we are working with strings

    load_special_indexes()

    if special_index_exists(model_class, field_name, index_type):
        return

    if environment.is_production_environment() or (
        in_testing() and not getattr(settings, "GENERATE_SPECIAL_INDEXES_DURING_TESTING", False)
    ):
        raise RuntimeError(
            "There is a missing index in your djangaeidx.yaml - \n\n{0}:\n\t{1}: [{2}]".format(
                _get_table_from_model(model_class), field_name, index_type
            )
        )

    _project_special_indexes.setdefault(
        _get_table_from_model(model_class), {}
    ).setdefault(field_name, []).append(str(index_type))

    write_special_indexes()


class Indexer(object):
    def handles(self, field, operator):
        """
            When given a field instance and an operator (e.g. gt, month__gt etc.)
            returns True or False whether or not this is the indexer to handle that
            situation
        """
        raise NotImplementedError()

    def validate_can_be_indexed(self, value, negated):
        """Return True if the value is indexable, False otherwise"""
        raise NotImplementedError()

    def prep_value_for_database(self, value, index): raise NotImplementedError()
    def prep_value_for_query(self, value): raise NotImplementedError()
    def indexed_column_name(self, field_column, value, index): raise NotImplementedError()
    def prep_query_operator(self, op):
        if "__" in op:
            return op.split("__")[-1]
        else:
            return "exact"  # By default do an exact operation

    def prepare_index_type(self, index_type, value): return index_type

    def unescape(self, value):
        value = value.replace("\\_", "_")
        value = value.replace("\\%", "%")
        value = value.replace("\\\\", "\\")
        return value


class StringIndexerMixin(object):
    STRING_FIELDS = (
        models.TextField,
        models.CharField,
        models.URLField,
        models.DateTimeField, # Django allows these for some reason
        models.DateField,
        models.TimeField,
        models.IntegerField, # SQL coerces ints to strings, and so we need these too
        models.PositiveIntegerField,
        models.AutoField
    )

    def handles(self, field, operator):
        try:
            # Make sure the operator is in there
            operator.split("__").index(self.OPERATOR)
        except ValueError:
            return False

        if isinstance(field, self.STRING_FIELDS):
            return True
        elif (
            isinstance(field, (iterable.ListField, iterable.SetField)) and
            field.item_field_type.__class__ in self.STRING_FIELDS and operator.startswith("item__")
        ):
            return True
        return False


class DateIndexerMixin(object):
    def handles(self, field, operator):
        DATE_FIELDS = (
            models.DateField,
            models.DateTimeField
        )

        if operator.split("__")[0] != self.OPERATOR:
            return False

        if isinstance(field, DATE_FIELDS):
            return True
        elif (
            isinstance(field, (iterable.ListField, iterable.SetField)) and
            field.item_field_type.__class__ in DATE_FIELDS and operator.startswith("item__")
        ):
            return True

        return False


class TimeIndexerMixin(object):
    def handles(self, field, operator):
        TIME_FIELDS = (
            models.TimeField,
            models.DateTimeField
        )

        if operator.split("__")[0] != self.OPERATOR:
            return False

        if isinstance(field, TIME_FIELDS):
            return True
        elif (
            isinstance(field, (iterable.ListField, iterable.SetField)) and
            field.item_field_type.__class__ in TIME_FIELDS and operator.startswith("item__")
        ):
            return True

        return False

class IExactIndexer(StringIndexerMixin, Indexer):
    OPERATOR = 'iexact'

    def validate_can_be_indexed(self, value, negated):
        return len(value) < 500

    def prep_value_for_database(self, value, index):
        if value is None:
            return None

        if isinstance(value, (int, long)):
            value = str(value)
        return value.lower()

    def prep_value_for_query(self, value):
        value = self.unescape(value)
        return value.lower()

    def indexed_column_name(self, field_column, value, index):
        return "_idx_iexact_{0}".format(field_column)


class HourIndexer(TimeIndexerMixin, Indexer):
    OPERATOR = 'hour'

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, datetime.datetime)

    def prep_value_for_database(self, value, index):
        if value:
            return value.hour
        return None

    def prep_value_for_query(self, value):
        if isinstance(value, (int, long)):
            return value

        if isinstance(value, basestring):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.hour

    def indexed_column_name(self, field_column, value, index):
        return "_idx_hour_{0}".format(field_column)


class MinuteIndexer(TimeIndexerMixin, Indexer):
    OPERATOR = 'minute'

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, datetime.datetime)

    def prep_value_for_database(self, value, index):
        if value:
            return value.minute
        return None

    def prep_value_for_query(self, value):
        if isinstance(value, (int, long)):
            return value

        if isinstance(value, basestring):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.minute

    def indexed_column_name(self, field_column, value, index):
        return "_idx_minute_{0}".format(field_column)


class SecondIndexer(TimeIndexerMixin, Indexer):
    OPERATOR = 'second'
    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, datetime.datetime)

    def prep_value_for_database(self, value, index):
        if value:
            return value.second
        return None

    def prep_value_for_query(self, value):
        if isinstance(value, (int, long)):
            return value

        if isinstance(value, basestring):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.second

    def indexed_column_name(self, field_column, value, index):
        return "_idx_second_{0}".format(field_column)


class DayIndexer(DateIndexerMixin, Indexer):
    OPERATOR = 'day'

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index):
        if value:
            return value.day
        return None

    def prep_value_for_query(self, value):
        if isinstance(value, (int, long)):
            return value

        if isinstance(value, basestring):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.day

    def indexed_column_name(self, field_column, value, index):
        return "_idx_day_{0}".format(field_column)


class YearIndexer(DateIndexerMixin, Indexer):
    OPERATOR = 'year'

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index):
        if value:
            return value.year
        return None

    def prep_value_for_query(self, value):
        if isinstance(value, (int, long)):
            return value

        if isinstance(value, basestring):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")

        return value.year

    def indexed_column_name(self, field_column, value, index):
        return "_idx_year_{0}".format(field_column)


class MonthIndexer(DateIndexerMixin, Indexer):
    OPERATOR = 'month'

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index):
        if value:
            return value.month
        return None

    def prep_value_for_query(self, value):
        if isinstance(value, (int, long)):
            return value

        if isinstance(value, basestring):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")

        return value.month

    def indexed_column_name(self, field_column, value, index):
        return "_idx_month_{0}".format(field_column)


class WeekDayIndexer(DateIndexerMixin, Indexer):
    OPERATOR = 'week_day'

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index):
        if value:
            zero_based_weekday = value.weekday()
            if zero_based_weekday == 6:  # Sunday
                return 1  # Django treats the week as starting at Sunday, but 1 based
            else:
                return zero_based_weekday + 2

        return None

    def prep_value_for_query(self, value):
        return value

    def indexed_column_name(self, field_column, value, index):
        return "_idx_week_day_{0}".format(field_column)


class ContainsIndexer(StringIndexerMixin, Indexer):
    OPERATOR = 'contains'

    def number_of_permutations(self, value):
        return sum(range(len(value)+1))

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False
        return isinstance(value, basestring) and len(value) <= 500

    def prep_value_for_database(self, value, index):
        results = []
        if value:
            # If this a date or a datetime, or something that supports isoformat, then use that
            if hasattr(value, "isoformat"):
                value = value.isoformat()

            if self.number_of_permutations(value) > MAX_COLUMNS_PER_SPECIAL_INDEX*500:
                raise ValueError(
                    "Can't index for contains query, this value is too long and has too many "
                    "permutations. You can increase the DJANGAE_MAX_COLUMNS_PER_SPECIAL_INDEX "
                    "setting to fix that. Use with caution."
                )
            if len(value) > CHARACTERS_PER_COLUMN[-1]:
                raise ValueError(
                    "Can't index for contains query, this value can be maximum {0} characters long."
                    .format(CHARACTERS_PER_COLUMN[-1])
                )

            if _is_iterable(value):
                # `value` is a list of strings. Generate a single combined list containing the
                # substrings of each string in `value`
                for element in value:
                    length = len(element)
                    lists = [element[i:j + 1] for i in xrange(length) for j in xrange(i, length)]
                    results.extend(lists)
            else:
                # `value` is a string. Generate a list of all its substrings.
                length = len(value)
                lists = [value[i:j + 1] for i in xrange(length) for j in xrange(i, length)]
                results.extend(lists)

        if not results:
            return None

        return _deduplicate_list(results)

    def prep_value_for_query(self, value):
        if hasattr(value, "isoformat"):
            value = value.isoformat()
        else:
            value = unicode(value)
        value = self.unescape(value)

        if STRIP_PERCENTS:
            # SQL does __contains by doing LIKE %value%
            if value.startswith("%") and value.endswith("%"):
                value = value[1:-1]

        return value

    def indexed_column_name(self, field_column, value, index):
        # This we use when we actually query to return the right field for a given
        # value length
        length = len(value)
        column_number = 0
        for x in CHARACTERS_PER_COLUMN:
            if length > x:
                column_number += 1
        return "_idx_contains_{0}_{1}".format(field_column, column_number)

    def prep_query_operator(self, op):
        return "exact"


class IContainsIndexer(ContainsIndexer):
    OPERATOR = 'icontains'

    def prep_value_for_database(self, value, index):
        if value is None:
            return None
        value = _make_lower(value)
        result = super(IContainsIndexer, self).prep_value_for_database(value, index)
        return result if result else None

    def indexed_column_name(self, field_column, value, index):
        column_name = super(IContainsIndexer, self).indexed_column_name(field_column, value, index)
        return column_name.replace('_idx_contains_', '_idx_icontains_')

    def prep_value_for_query(self, value):
        return super(IContainsIndexer, self).prep_value_for_query(value).lower()


class EndsWithIndexer(StringIndexerMixin, Indexer):
    """
        dbindexer originally reversed the string and did a startswith on it.
        However, this is problematic as it uses an inequality and therefore
        limits the queries you can perform. Instead, we store all permutations
        of the last characters in a list field. Then we can just do an exact lookup on
        the value. Which isn't as nice, but is more flexible.
    """
    OPERATOR = 'endswith'

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False

        return isinstance(value, basestring) and len(value) < 500

    def prep_value_for_database(self, value, index):
        if value is None:
            return None

        results = []
        if _is_iterable(value):
            # `value` is a list of strings. Create a single combined list of "endswith" values
            # of all the strings in the list
            for element in value:
                for i in xrange(0, len(element)):
                    results.append(element[i:])
        else:
            # `value` is a string. Create a list of "endswith" strings.
            for i in xrange(0, len(value)):
                results.append(value[i:])

        if not results:
            return None

        return _deduplicate_list(results)

    def prep_value_for_query(self, value):
        value = self.unescape(value)
        if STRIP_PERCENTS:
            if value.startswith("%"):
                value = value[1:]
        return value

    def indexed_column_name(self, field_column, value, index):
        return "_idx_endswith_{0}".format(field_column)

    def prep_query_operator(self, op):
        return "exact"


class IEndsWithIndexer(EndsWithIndexer):
    """
        Same as above, just all lower cased
    """
    OPERATOR = 'iendswith'

    def prep_value_for_database(self, value, index):
        if value:
            value = _make_lower(value)
        return super(IEndsWithIndexer, self).prep_value_for_database(value, index)

    def prep_value_for_query(self, value):
        return super(IEndsWithIndexer, self).prep_value_for_query(value.lower())

    def indexed_column_name(self, field_column, value, index):
        return "_idx_iendswith_{0}".format(field_column)


class StartsWithIndexer(StringIndexerMixin, Indexer):
    """
        Although we can do a startswith natively, doing it this way allows us to
        use more queries (E.g. we save an exclude)
    """
    OPERATOR = 'startswith'

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False

        return isinstance(value, basestring) and len(value) < 500

    def prep_value_for_database(self, value, index):
        if value is None:
            return None

        if isinstance(value, datetime.datetime):
            value = value.strftime("%Y-%m-%d %H:%M:%S")

        results = []
        if _is_iterable(value):
            # `value` is a list of strings. Create a single combined list of "startswith" values
            # of all the strings in the list
            for element in value:
                for i in xrange(1, len(element) + 1):
                    results.append(element[:i])
        else:
            # `value` is a string. Create a list of "startswith" strings.
            for i in xrange(1, len(value) + 1):
                results.append(value[:i])

        if not results:
            return None

        return _deduplicate_list(results)

    def prep_value_for_query(self, value):
        value = self.unescape(value)
        if STRIP_PERCENTS:
            if value.endswith("%"):
                value = value[:-1]
        return value

    def indexed_column_name(self, field_column, value, index):
        return "_idx_startswith_{0}".format(field_column)

    def prep_query_operator(self, op):
        return "exact"


class IStartsWithIndexer(StartsWithIndexer):
    """
        Same as above, just all lower cased
    """
    OPERATOR = 'istartswith'

    def prep_value_for_database(self, value, index):
        if value:
            value = _make_lower(value)
        return super(IStartsWithIndexer, self).prep_value_for_database(value, index)

    def prep_value_for_query(self, value):
        return super(IStartsWithIndexer, self).prep_value_for_query(value.lower())

    def indexed_column_name(self, field_column, value, index):
        return "_idx_istartswith_{0}".format(field_column)


class RegexIndexer(StringIndexerMixin, Indexer):
    OPERATOR = 'regex'

    def prepare_index_type(self, index_type, value):
        """
            If we're dealing with RegexIndexer, we create a new index for each
            regex pattern. Indexes are called regex__pattern.
        """
        return '{}__{}'.format(index_type, value.encode("utf-8").encode('hex'))

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False

        return isinstance(value, bool)

    def get_pattern(self, index):
        try:
            return index.split('__')[-1].decode('hex').decode("utf-8")
        except IndexError:
            return ''

    def check_if_match(self, value, index, flags=0):
        pattern = self.get_pattern(index)

        if value:
            if _is_iterable(value):
                if any([bool(re.search(pattern, x, flags)) for x in value]):
                    return True
            else:
                if isinstance(value, (int, long)):
                    value = str(value)

                return bool(re.search(pattern, value, flags))
        return False

    def prep_value_for_database(self, value, index):
        return self.check_if_match(value, index)

    def prep_value_for_query(self, value):
        return True

    def indexed_column_name(self, field_column, value, index):
        return "_idx_regex_{0}_{1}".format(
            field_column, self.get_pattern(index).encode("utf-8").encode('hex')
        )

    def prep_query_operator(self, op):
        return "exact"


class IRegexIndexer(RegexIndexer):
    OPERATOR = 'iregex'

    def prepare_index_type(self, index_type, value):
        return '{}__{}'.format(index_type, value.encode('hex'))

    def prep_value_for_database(self, value, index):
        return self.check_if_match(value, index, flags=re.IGNORECASE)

    def indexed_column_name(self, field_column, value, index):
        return "_idx_iregex_{0}_{1}".format(field_column, self.get_pattern(index).encode('hex'))


_REGISTERED_INDEXERS = []

def register_indexer(indexer_class):
    global _REGISTERED_INDEXERS
    _REGISTERED_INDEXERS.append(indexer_class())


def get_indexer(field, operator):
    global _REGISTERED_INDEXERS

    for indexer in _REGISTERED_INDEXERS:
        if indexer.handles(field, operator):
            return indexer


register_indexer(IExactIndexer)
register_indexer(ContainsIndexer)
register_indexer(IContainsIndexer)
register_indexer(HourIndexer)
register_indexer(MinuteIndexer)
register_indexer(SecondIndexer)
register_indexer(DayIndexer)
register_indexer(MonthIndexer)
register_indexer(YearIndexer)
register_indexer(WeekDayIndexer)
register_indexer(EndsWithIndexer)
register_indexer(IEndsWithIndexer)
register_indexer(StartsWithIndexer)
register_indexer(IStartsWithIndexer)
register_indexer(RegexIndexer)
register_indexer(IRegexIndexer)



# LIBRARIES
from contextlib import contextmanager

# DJANGAE
from djangae.contrib import sleuth
from djangae.core.management import execute_from_command_line
from djangae.test import TestCase


@contextmanager
def test_context_manager(*args, **kwargs):
    yield

class ManagementCommandsTest(TestCase):
    def test_arguments_are_passed_through_correctly(self):
        with sleuth.switch("django.core.management.execute_from_command_line", lambda *args, **kwargs: None) as django_execute_mock, \
             sleuth.switch("djangae.sandbox.activate", test_context_manager):
            execute_from_command_line(['manage.py', 'arg1', 'arg2', 'arg3',])
            self.assertEqual(1, django_execute_mock.call_count)
            self.assertEqual((['manage.py', 'arg1', 'arg2', 'arg3',],), django_execute_mock.calls[0].args)

    def test_sandbox_can_be_specified(self):
        with sleuth.switch("django.core.management.execute_from_command_line", lambda *args, **kwargs: None) as django_execute_mock, \
             sleuth.switch("djangae.sandbox.activate", test_context_manager) as activate_sandbox_mock:

            # test default sandbox is used if no sandbox argument
            execute_from_command_line(['manage.py', 'arg1', 'arg2',])
            self.assertEqual(1, activate_sandbox_mock.call_count)
            self.assertEqual('local', activate_sandbox_mock.calls[0].args[0])
            self.assertEqual(1, django_execute_mock.call_count)
            self.assertEqual((['manage.py', 'arg1', 'arg2',],), django_execute_mock.calls[0].args)

            # test that sandbox argument is used when given
            execute_from_command_line(['manage.py', '--sandbox=test', 'arg1', 'arg2',])
            self.assertEqual(2, activate_sandbox_mock.call_count)
            self.assertEqual('test', activate_sandbox_mock.calls[1].args[0])
            self.assertEqual(2, django_execute_mock.call_count)
            self.assertEqual((['manage.py', 'arg1', 'arg2',],), django_execute_mock.calls[1].args)

            # test that sandbox argument can be in any position
            execute_from_command_line(['manage.py', 'arg1', '--sandbox=test', 'arg2',])
            self.assertEqual(3, activate_sandbox_mock.call_count)
            self.assertEqual('test', activate_sandbox_mock.calls[2].args[0])
            self.assertEqual(3, django_execute_mock.call_count)
            self.assertEqual((['manage.py', 'arg1', 'arg2',],), django_execute_mock.calls[2].args)



import os
from setuptools import setup, find_packages


NAME = 'djangae'
PACKAGES = find_packages()
DESCRIPTION = 'Django integration with Google App Engine'
URL = "https://github.com/potatolondon/djangae"
LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()
AUTHOR = 'Potato London Ltd.'

EXTRAS = {
    "test": ["webtest"],
}

setup(
    name=NAME,
    version='0.9.9',
    packages=PACKAGES,

    # metadata for upload to PyPI
    author=AUTHOR,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    keywords=["django", "Google App Engine", "GAE"],
    url=URL,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
#        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],

    include_package_data=True,
    # dependencies
    extras_require=EXTRAS,
    tests_require=EXTRAS['test'],
)
